import logging
import numpy as np
import scipy.ndimage as spim
import scipy.spatial as sptl
from scipy import fft as sp_ft
from skimage.measure import regionprops
from deprecated import deprecated
from porespy.tools import extend_slice
from porespy.tools import _check_for_singleton_axes
from porespy.tools import Results
from porespy import settings
from porespy.tools import get_tqdm
from numba import njit


__all__ = [
    "boxcount",
    "chord_counts",
    "chord_length_distribution",
    "find_h",
    "lineal_path_distribution",
    "pore_size_distribution",
    "radial_density_distribution",
    "porosity",
    "porosity_profile",
    "representative_elementary_volume",
    "satn_profile",
    "two_point_correlation",
    "phase_fraction",
    "pc_curve",
    "pc_curve_from_ibip",
    "pc_curve_from_mio",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


def boxcount(im, bins=10):
    r"""
    Calculates fractal dimension of an image using the tiled box counting
    method [1]_

    Parameters
    ----------
    im : ndarray
        The image of the porous material.
    bins : int or array_like, optional
        The number of box sizes to use. The default is 10 sizes
        logarithmically spaced between 1 and ``min(im.shape)``.
        If an array is provided, this is used directly.

    Returns
    -------
    results
        An object possessing the following attributes:

        size : ndarray
            The box sizes used
        count : ndarray
            The number of boxes of each size that contain both solid and void
        slope : ndarray
            The gradient of ``count``. This has the same number of elements as
            ``count`` and

    References
    ----------
    .. [1] See Boxcounting on `Wikipedia <https://en.wikipedia.org/wiki/Box_counting>`_

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/box_counting.html>`_
    to view online example.

    """
    im = np.array(im, dtype=bool)

    if (len(im.shape) != 2 and len(im.shape) != 3):
        raise Exception('Image must be 2-dimensional or 3-dimensional')

    if isinstance(bins, int):
        Ds = np.unique(np.logspace(1, np.log10(min(im.shape)), bins).astype(int))
    else:
        Ds = np.array(bins).astype(int)

    N = []
    for d in tqdm(Ds, **settings.tqdm):
        result = 0
        for i in range(0, im.shape[0], d):
            for j in range(0, im.shape[1], d):
                if len(im.shape) == 2:
                    temp = im[i:i+d, j:j+d]
                    result += np.any(temp)
                    result -= np.all(temp)
                else:
                    for k in range(0, im.shape[2], d):
                        temp = im[i:i+d, j:j+d, k:k+d]
                        result += np.any(temp)
                        result -= np.all(temp)
        N.append(result)
    slope = -1*np.gradient(np.log(np.array(N)), np.log(Ds))
    data = Results()
    data.size = Ds
    data.count = N
    data.slope = slope
    return data


def representative_elementary_volume(im, npoints=1000):
    r"""
    Calculates the porosity of an image as a function subdomain size.

    This function extracts a specified number of subdomains of random size,
    then finds their porosity.

    Parameters
    ----------
    im : ndarray
        The image of the porous material
    npoints : int
        The number of randomly located and sized boxes to sample.  The default
        is 1000.

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        'volume'
            The total volume of each cubic subdomain tested
        'porosity'
            The porosity of each subdomain tested

        These attributes can be conveniently plotted by passing the Results
        object to matplotlib's ``plot`` function using the
        \* notation: ``plt.plot(\*result, 'b.')``.  The resulting plot is
        similar to the sketch given by Bachmat and Bear [1]_

    Notes
    -----
    This function is frustratingly slow.  Profiling indicates that all the time
    is spent on scipy's ``sum`` function which is needed to sum the number of
    void voxels (1's) in each subdomain.

    References
    ----------
    .. [1] Bachmat and Bear. On the Concept and Size of a Representative
       Elementary Volume (Rev), Advances in Transport Phenomena in Porous Media
       (1987)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/representative_elementary_volume.html>`_
    to view online example.

    """
    # TODO: this function is a prime target for parallelization since the
    # ``npoints`` are calculated independenlty.
    im_temp = np.zeros_like(im)
    crds = np.array(np.random.rand(npoints, im.ndim) * im.shape, dtype=int)
    pads = np.array(np.random.rand(npoints) * np.amin(im.shape) / 2 + 10, dtype=int)
    im_temp[tuple(crds.T)] = True
    labels, N = spim.label(input=im_temp)
    slices = spim.find_objects(input=labels)
    porosity = np.zeros(shape=(N,), dtype=float)
    volume = np.zeros(shape=(N,), dtype=int)
    for i in tqdm(np.arange(0, N), **settings.tqdm):
        s = slices[i]
        p = pads[i]
        new_s = extend_slice(s, shape=im.shape, pad=p)
        temp = im[new_s]
        Vp = np.sum(temp, dtype=np.int64)
        Vt = np.size(temp)
        porosity[i] = Vp / Vt
        volume[i] = Vt
    profile = Results()
    profile.volume = volume
    profile.porosity = porosity
    return profile


def porosity(im):
    r"""
    Calculates the porosity of an image assuming 1's are void space and 0's
    are solid phase.

    All other values are ignored, so this can also return the relative
    fraction of a phase of interest in multiphase images.

    Parameters
    ----------
    im : ndarray
        Image of the void space with 1's indicating void phase (or ``True``)
        and 0's indicating the solid phase (or ``False``). All other values
        are ignored (see Notes).

    Returns
    -------
    porosity : float
        Calculated as the sum of all 1's divided by the sum of all 1's and 0's.

    See Also
    --------
    phase_fraction
    find_outer_region

    Notes
    -----
    This function assumes void is represented by 1 and solid by 0, and all
    other values are ignored.  This is useful, for example, for images of
    cylindrical cores, where all voxels outside the core are labelled with 2.

    Alternatively, images can be processed with ``find_disconnected_voxels``
    to get an image of only blind pores.  This can then be added to the orignal
    image such that blind pores have a value of 2, thus allowing the
    calculation of accessible porosity, rather than overall porosity.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/porosity.html>`_
    to view online example.

    """
    im = np.array(im, dtype=np.int64)
    Vp = np.sum(im == 1, dtype=np.int64)
    Vs = np.sum(im == 0, dtype=np.int64)
    e = Vp / (Vs + Vp)
    return e


def porosity_profile(im, axis=0):
    r"""
    Computes the porosity profile along the specified axis

    Parameters
    ----------
    im : ndarray
        The volumetric image for which to calculate the porosity profile.  All
        voxels with a value of 1 (or ``True``) are considered as void.
    axis : int
        The axis (0, 1, or 2) along which to calculate the profile.  For
        instance, if `axis` is 0, then the porosity in each YZ plane is
        calculated and returned as 1D array with 1 value for each X position.

    Returns
    -------
    result : 1D-array
        A 1D-array of porosity along the specified axis

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/porosity_profile.html>`_
    to view online example.

    """
    if axis >= im.ndim:
        raise Exception('axis out of range')
    im = np.atleast_3d(im)
    a = set(range(im.ndim)).difference(set([axis]))
    a1, a2 = a
    tmp = np.sum(np.sum(im == 1, axis=a2, dtype=np.int64), axis=a1, dtype=np.int64)
    prof = tmp / (im.shape[a2] * im.shape[a1])
    return prof


def radial_density_distribution(dt, bins=10, log=False, voxel_size=1):
    r"""
    Computes radial density function by analyzing the histogram of voxel
    values in the distance transform.  This function is defined by
    Torquato [1] as:

        .. math::

            \int_0^\infty P(r)dr = 1.0

    where *P(r)dr* is the probability of finding a voxel at a lying at a radial
    distance between *r* and *dr* from the solid interface.  This is equivalent
    to a probability density function (*pdf*)

    The cumulative distribution is defined as:

        .. math::

            F(r) = \int_r^\infty P(r)dr

    which gives the fraction of pore-space with a radius larger than *r*. This
    is equivalent to the cumulative distribution function (*cdf*).

    Parameters
    ----------
    dt : ndarray
        A distance transform of the pore space (the ``edt`` package is
        recommended).  Note that it is recommended to apply
        ``find_dt_artifacts`` to this image first, and set potentially
        erroneous values to 0 with ``dt[mask] = 0`` where
        ``mask = porespy.filters.find_dt_artifaces(dt)``.
    bins : int or array_like
        This number of bins (if int) or the location of the bins (if array).
        This argument is passed directly to Scipy's ``histogram`` function so
        see that docstring for more information.  The default is 10 bins, which
        reduces produces a relatively smooth distribution.
    log : boolean
        If ``True`` the size data is converted to log (base-10)
        values before processing.  This can help to plot wide size
        distributions or to better visualize the in the small size region.
        Note that you should not anti-log the radii values in the retunred
        ``tuple``, since the binning is performed on the logged radii values.
    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so the
        user can apply the scaling to the returned results after the fact.

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        ============== =======================================================
        Attribute      Description
        ============== =======================================================
        *R* or *LogR*  Radius, equivalent to ``bin_centers``
        *pdf*          Probability density function
        *cdf*          Cumulative density function
        *bin_centers*  The center point of each bin
        *bin_edges*    Locations of bin divisions, including 1 more value than
                       the number of bins
        *bin_widths*   Useful for passing to the ``width`` argument of
                       ``matplotlib.pyplot.bar``
        ============== =======================================================

    Notes
    -----
    Torquato refers to this as the *pore-size density function*, and mentions
    that it is also known as the *pore-size distribution function*.  These
    terms are avoided here since they have specific connotations in porous
    media analysis.

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002) - See page 48 & 292

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/radial_density.html>`_
    to view online example.

    """
    im = np.copy(dt)
    x = im[im > 0].flatten()
    if log:
        x = np.log10(x)
    h = np.histogram(x, bins=bins, density=True)
    h = _parse_histogram(h=h, voxel_size=voxel_size)
    rdf = Results()
    rdf[f"{log*'Log' + 'R'}"] = h.bin_centers
    rdf.pdf = h.pdf
    rdf.cdf = h.cdf
    rdf.relfreq = h.relfreq
    rdf.bin_centers = h.bin_centers
    rdf.bin_edges = h.bin_edges
    rdf.bin_widths = h.bin_widths
    return rdf


def lineal_path_distribution(im, bins=10, voxel_size=1, log=False):
    r"""
    Determines the probability that a point lies within a certain distance
    of the opposite phase *along a specified direction*

    This relates directly the radial density function defined by Torquato [1],
    but instead of reporting the probability of lying within a stated distance
    to the nearest solid in any direciton, it considers only linear distances
    along orthogonal directions.The benefit of this is that anisotropy can be
    detected in materials by performing the analysis in multiple orthogonal
    directions.

    Parameters
    ----------
    im : ndarray
        An image with each voxel containing the distance to the nearest solid
        along a linear path, as produced by ``distance_transform_lin``.
    bins : int or array_like
        The number of bins or a list of specific bins to use
    voxel_size : scalar
        The side length of a voxel.  This is used to scale the chord lengths
        into real units.  Note this is applied *after* the binning, so
        ``bins``, if supplied, should be in terms of voxels, not length units.
    log : boolean
        If ``True`` (default) the size data is converted to log (base-10)
        values before processing.  This can help to plot wide size
        distributions or to better visualize data in the small size region.
        Note that you should not anti-log the radii values in the retunred
        ``results``, since the binning is performed on the logged radii values.

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        *L* or *LogL*
            Length, equivalent to ``bin_centers``
        *pdf*
            Probability density function
        *cdf*
            Cumulative density function
        *relfreq*
            Relative frequency chords in each bin.  The sum of all bin
            heights is 1.0.  For the cumulative relativce, use *cdf* which is
            already normalized to 1.
        *bin_centers*
            The center point of each bin
        *bin_edges*
            Locations of bin divisions, including 1 more value than
            the number of bins
        *bin_widths*
            Useful for passing to the ``width`` argument of
            ``matplotlib.pyplot.bar``

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/linearl_path_distribution.html>`_
    to view online example.

    """
    x = im[im > 0]
    if log:
        x = np.log10(x)
    h = list(np.histogram(x, bins=bins, density=True))
    h = _parse_histogram(h=h, voxel_size=voxel_size)
    cld = Results()
    cld[f"{log*'Log' + 'L'}"] = h.bin_centers
    cld.pdf = h.pdf
    cld.cdf = h.cdf
    cld.relfreq = h.relfreq
    cld.bin_centers = h.bin_centers
    cld.bin_edges = h.bin_edges
    cld.bin_widths = h.bin_widths
    return cld


def chord_length_distribution(im, bins=10, log=False, voxel_size=1,
                              normalization='count'):
    r"""
    Determines the distribution of chord lengths in an image containing chords.

    Parameters
    ----------
    im : ndarray
        An image with chords drawn in the pore space, as produced by
        ``apply_chords`` or ``apply_chords_3d``.  ``im`` can be either boolean,
        in which case each chord will be identified using ``scipy.ndimage.label``,
        or numerical values in case it is assumed that chords have already been
        identifed and labeled. In both cases, the size of each chord will be
        computed as the number of voxels belonging to each labelled region.
    bins : scalar or array_like
        If a scalar is given it is interpreted as the number of bins to use,
        and if an array is given they are used as the bins directly.
    log : boolean
        If ``True`` (default) the size data is converted to log (base-10)
        values before processing.  This can help to plot wide size
        distributions or to better visualize the in the small size region.
        Note that you should not anti-log the radii values in the retunred
        ``tuple``, since the binning is performed on the logged radii values.
    normalization : string
        Indicates how to normalize the bin heights.  Options are:

        *'count' or 'number'*
            (default) This simply counts the number of chords in each bin in
            the normal sense of a histogram.  This is the rigorous definition
            according to Torquato [1].
        *'length'*
            This multiplies the number of chords in each bin by the
            chord length (i.e. bin size).  The normalization scheme accounts for
            the fact that long chords are less frequent than shorert chords,
            thus giving a more balanced distribution.

    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so the
        user can apply the scaling to the returned results after the fact.

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        *L* or *LogL*
            Chord length, equivalent to ``bin_centers``
        *pdf*
            Probability density function
        *cdf*
            Cumulative density function
        *relfreq*
            Relative frequency chords in each bin.  The sum of all bin
            heights is 1.0.  For the cumulative relativce, use *cdf* which is
            already normalized to 1.
        *bin_centers*
            The center point of each bin
        *bin_edges*
            Locations of bin divisions, including 1 more value than
            the number of bins
        *bin_widths*
            Useful for passing to the ``width`` argument of
            ``matplotlib.pyplot.bar``

    References
    ----------
    [1] Torquato, S. Random Heterogeneous Materials: Mircostructure and
    Macroscopic Properties. Springer, New York (2002) - See page 45 & 292

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/chord_length_distribution.html>`_
    to view online example.

    """
    x = chord_counts(im)
    if bins is None:
        bins = np.array(range(0, x.max() + 2)) * voxel_size
    x = x * voxel_size
    if log:
        x = np.log10(x)
    if normalization == 'length':
        h = list(np.histogram(x, bins=bins, density=False))
        # Scale bin heigths by length
        h[0] = h[0] * (h[1][1:] + h[1][:-1]) / 2
        # Normalize h[0] manually
        h[0] = h[0] / h[0].sum(dtype=np.int64) / (h[1][1:] - h[1][:-1])
    elif normalization in ['number', 'count']:
        h = np.histogram(x, bins=bins, density=True)
    else:
        raise Exception('Unsupported normalization:', normalization)
    h = _parse_histogram(h)
    cld = Results()
    cld[f"{log*'Log' + 'L'}"] = h.bin_centers
    cld.pdf = h.pdf
    cld.cdf = h.cdf
    cld.relfreq = h.relfreq
    cld.bin_centers = h.bin_centers
    cld.bin_edges = h.bin_edges
    cld.bin_widths = h.bin_widths
    return cld


def pore_size_distribution(im, bins=10, log=True, voxel_size=1):
    r"""
    Calculate a pore-size distribution based on the image produced by the
    ``porosimetry`` or ``local_thickness`` functions.

    Parameters
    ----------
    im : ndarray
        The array of containing the sizes of the largest sphere that overlaps
        each voxel.  Obtained from either ``porosimetry`` or
        ``local_thickness``.
    bins : scalar or array_like
        Either an array of bin sizes to use, or the number of bins that should
        be automatically generated that span the data range.
    log : boolean
        If ``True`` (default) the size data is converted to log (base-10)
        values before processing.  This can help to plot wide size
        distributions or to better visualize the in the small size region.
        Note that you should not anti-log the radii values in the retunred
        ``tuple``, since the binning is performed on the logged radii values.
    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so the
        user can apply the scaling to the returned results after the fact.

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        *R* or *logR*
            Radius, equivalent to ``bin_centers``
        *pdf*
            Probability density function
        *cdf*
            Cumulative density function
        *satn*
            Phase saturation in differential form.  For the cumulative
            saturation, just use *cfd* which is already normalized to 1.
        *bin_centers*
            The center point of each bin
        *bin_edges*
            Locations of bin divisions, including 1 more value than
            the number of bins
        *bin_widths*
            Useful for passing to the ``width`` argument of
            ``matplotlib.pyplot.bar``

    Notes
    -----
    (1) To ensure the returned values represent actual sizes you can manually
    scale the input image by the voxel size first (``im *= voxel_size``)

    plt.bar(psd.R, psd.satn, width=psd.bin_widths, edgecolor='k')

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/pore_size_distribution.html>`_
    to view online example.

    """
    im = im.flatten()
    vals = im[im > 0] * voxel_size
    if log:
        vals = np.log10(vals)
    h = _parse_histogram(np.histogram(vals, bins=bins, density=True))
    cld = Results()
    cld[f"{log*'Log' + 'R'}"] = h.bin_centers
    cld.pdf = h.pdf
    cld.cdf = h.cdf
    cld.satn = h.relfreq
    cld.bin_centers = h.bin_centers
    cld.bin_edges = h.bin_edges
    cld.bin_widths = h.bin_widths
    return cld


def two_point_correlation_bf(im, spacing=10):
    r"""
    Calculates the two-point correlation function using brute-force (see Notes)

    Parameters
    ----------
    im : ndarray
        The image of the void space on which the 2-point correlation is
        desired.
    spacing : int
        The space between points on the regular grid that is used to
        generate the correlation (see Notes).

    Returns
    -------
    result : Results object
        A custom object with the following data added as named attributes:

        'distance'
            The distance between two points. The distance values are binned
            as:
        $$ bins = range(start=0, stop=np.amin(im.shape)/2, stride=spacing) $$

        'probability'
            The probability that two points of the stated separation distance
            are within the same phase

    Notes
    -----
    The brute-force approach means overlaying a grid of equally spaced points
    onto the image, calculating the distance between each and every pair of
    points, then counting the instances where both pairs lie in the void space.

    This approach uses a distance matrix so can consume memory very quickly for
    large 3D images and/or close spacing.  It is recommended to avoid this.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/two_point_correlation_bf.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)
    if im.ndim == 2:
        pts = np.meshgrid(range(0, im.shape[0], spacing),
                          range(0, im.shape[1], spacing))
        crds = np.vstack([pts[0].flatten(),
                          pts[1].flatten()]).T
    elif im.ndim == 3:
        pts = np.meshgrid(range(0, im.shape[0], spacing),
                          range(0, im.shape[1], spacing),
                          range(0, im.shape[2], spacing))
        crds = np.vstack([pts[0].flatten(),
                          pts[1].flatten(),
                          pts[2].flatten()]).T
    dmat = sptl.distance.cdist(XA=crds, XB=crds)
    hits = im[tuple(pts)].flatten()
    dmat = dmat[hits, :]
    h1 = np.histogram(dmat, bins=range(0, int(np.amin(im.shape) / 2), spacing))
    dmat = dmat[:, hits]
    h2 = np.histogram(dmat, bins=h1[1])
    tpcf = Results()
    tpcf.distance = h2[1][:-1]
    tpcf.probability = h2[0] / h1[0]
    return tpcf


def _radial_profile(autocorr, bins, pf=None, voxel_size=1):
    r"""
    Helper functions to calculate the radial profile of the autocorrelation

    Masks the image in radial segments from the center and averages the values
    The distance values are normalized and 100 bins are used as default.

    Parameters
    ----------
    autocorr : ndarray
        The image of autocorrelation produced by FFT
    r_max : int or float
        The maximum radius in pixels to sum the image over
    bins : ndarray
        The edges of the bins to use in summing the radii, ** must be in voxels
    pf : float
        the phase fraction (porosity) of the image, used for scaling the
        normalized autocorrelation down to match the two-point correlation
        definition as given by Torquato
    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so the
        user can apply the scaling to the returned results after the fact.

    Returns
    -------
    result : tpcf


    """
    if len(autocorr.shape) == 2:
        adj = np.reshape(autocorr.shape, [2, 1, 1])
        # use np.round otherwise with odd image sizes, the mask generated can
        # be zero, resulting in Div/0 error
        inds = np.indices(autocorr.shape) - np.round(adj / 2)
        dt = np.sqrt(inds[0]**2 + inds[1]**2)
    elif len(autocorr.shape) == 3:
        adj = np.reshape(autocorr.shape, [3, 1, 1, 1])
        # use np.round otherwise with odd image sizes, the mask generated can
        # be zero, resulting in Div/0 error
        inds = np.indices(autocorr.shape) - np.round(adj / 2)
        dt = np.sqrt(inds[0]**2 + inds[1]**2 + inds[2]**2)
    else:
        raise Exception('Image dimensions must be 2 or 3')
    if np.max(bins) > np.max(dt):
        msg = (
            'Bins specified distances exceeding maximum radial distance for'
            ' image size. Radial distance cannot exceed distance from center'
            ' of image to corner.'
        )
        raise Exception(msg)

    bin_size = bins[1:] - bins[:-1]
    radial_sum = _get_radial_sum(dt, bins, bin_size, autocorr)
    # Return normalized bin and radially summed autoc
    norm_autoc_radial = radial_sum / np.max(autocorr)
    h = [norm_autoc_radial, bins]
    h = _parse_histogram(h, voxel_size=1)
    tpcf = Results()
    tpcf.distance = h.bin_centers * voxel_size
    tpcf.bin_centers = h.bin_centers * voxel_size
    tpcf.bin_edges = h.bin_edges * voxel_size
    tpcf.bin_widths = h.bin_widths * voxel_size
    tpcf.probability = norm_autoc_radial
    tpcf.probability_scaled = norm_autoc_radial * pf
    tpcf.pdf = h.pdf * pf
    tpcf.relfreq = h.relfreq
    return tpcf


@njit(parallel=False)
def _get_radial_sum(dt, bins, bin_size, autocorr):
    radial_sum = np.zeros_like(bins[:-1])
    for i, r in enumerate(bins[:-1]):
        mask = (dt <= r) * (dt > (r - bin_size[i]))
        radial_sum[i] = np.sum(np.ravel(autocorr)[np.ravel(mask)], dtype=np.int64) \
            / np.sum(mask)
    return radial_sum


def two_point_correlation(im, voxel_size=1, bins=100):
    r"""
    Calculate the two-point correlation function using Fourier transforms

    Parameters
    ----------
    im : ndarray
        The image of the void space on which the 2-point correlation is
        desired, in which the phase of interest is labelled as True
    voxel_size : scalar
        The size of a voxel side in preferred units.  The default is 1, so
        the user can apply the scaling to the returned results after the
        fact.
    bins : scalar or array_like
        Either an array of bin sizes to use, or the number of bins that
        should be automatically generated that span the data range. The
        maximum value of the bins, if passed as an array, cannot exceed
        the distance from the center of the image to the corner.

    Returns
    -------
    result : tpcf
        The two-point correlation function object, with named attributes:

        *distance*
            The distance between two points, equivalent to bin_centers
        *bin_centers*
            The center point of each bin. See distance
        *bin_edges*
            Locations of bin divisions, including 1 more value than
            the number of bins
        *bin_widths*
            Useful for passing to the ``width`` argument of
            ``matplotlib.pyplot.bar``
        *probability_normalized*
            The probability that two points of the stated separation distance
            are within the same phase normalized to 1 at r = 0
        *probability* or *pdf*
            The probability that two points of the stated separation distance
            are within the same phase scaled to the phase fraction at r = 0

    Notes
    -----
    The fourier transform approach utilizes the fact that the
    autocorrelation function is the inverse FT of the power spectrum
    density. For background read the Scipy fftpack docs and for a good
    explanation `see this thesis
    <https://www.ucl.ac.uk/~ucapikr/projects/KamilaSuankulova_BSc_Project.pdf>`_.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/two_point_correlation.html>`_
    to view online example.

    """
    # Get the number of CPUs available to parallel process Fourier transforms
    cpus = settings.ncores
    # Get the phase fraction of the image
    pf = porosity(im)
    if isinstance(bins, int):
        # Calculate half lengths of the image
        r_max = (np.ceil(np.min(np.shape(im))) / 2).astype(int)
        # Get the bin-size - ensures it will be at least 1
        bin_size = int(np.ceil(r_max / bins))
        # Calculate the bin divisions, equivalent to bin_edges
        bins = np.arange(0, r_max + bin_size, bin_size)
    # set the number of parallel processors to use:
    with sp_ft.set_workers(cpus):
        # Fourier Transform and shift image
        F = sp_ft.ifftshift(sp_ft.rfftn(sp_ft.fftshift(im)))
        # Compute Power Spectrum
        P = np.absolute(F**2)
        # Auto-correlation is inverse of Power Spectrum
        autoc = np.absolute(sp_ft.ifftshift(sp_ft.irfftn(sp_ft.fftshift(P))))
    tpcf = _radial_profile(autoc, bins, pf=pf, voxel_size=voxel_size)
    return tpcf


def _parse_histogram(h, voxel_size=1, density=True):
    delta_x = h[1]
    P = h[0]
    bin_widths = delta_x[1:] - delta_x[:-1]
    temp = P * (bin_widths)
    C = np.cumsum(temp[-1::-1])[-1::-1]
    S = P * (bin_widths)
    if not density:
        P /= np.max(P)
        temp_sum = np.sum(P * bin_widths)
        C /= temp_sum
        S /= temp_sum

    bin_edges = delta_x * voxel_size
    bin_widths = (bin_widths) * voxel_size
    bin_centers = ((delta_x[1:] + delta_x[:-1]) / 2) * voxel_size
    hist = Results()
    hist.pdf = P
    hist.cdf = C
    hist.relfreq = S
    hist.bin_centers = bin_centers
    hist.bin_edges = bin_edges
    hist.bin_widths = bin_widths
    return hist


def chord_counts(im):
    r"""
    Find the length of each chord in the supplied image

    Parameters
    ----------
    im : ndarray
        An image containing chords drawn in the void space.

    Returns
    -------
    result : 1D-array
        A 1D array with one element for each chord, containing its length.

    Notes
    ----
    The returned array can be passed to ``plt.hist`` to plot the histogram,
    or to ``np.histogram`` to get the histogram data directly. Another useful
    function is ``np.bincount`` which gives the number of chords of each
    length in a format suitable for ``plt.plot``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/reference/metrics/chord_counts.html>`_
    to view online example.

    """
    labels, N = spim.label(im > 0)
    props = regionprops(labels)
    chord_lens = np.array([i.filled_area for i in props], dtype=int)
    return chord_lens


def phase_fraction(im, normed=True):
    r"""
    Calculate the fraction of each phase in an image

    Parameters
    ----------
    im : ndarray
        An ndarray containing integer values
    normed : boolean
        If ``True`` (default) the returned values are normalized by the total
        number of voxels in image, otherwise the voxel count of each phase is
        returned.

    Returns
    -------
    result : 1D-array
        A array of length max(im) with each element containing the number of
        voxels found with the corresponding label.

    See Also
    --------
    porosity

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/phase_fraction.html>`_
    to view online example.

    """
    if im.dtype == bool:
        im = im.astype(int)
    labels = np.unique(im)
    results = {}
    for label in labels:
        results[label] = np.sum(im == label, dtype=np.int64) * \
            (1 / im.size if normed else 1)
    return results


@deprecated("This function is deprecated, use pc_curve instead")
def pc_curve_from_ibip(*args, **kwargs):
    r"""
    This function is deprecated.  Use ``pc_curve`` instead.  Note that the
    ``stepped`` argument is no longer supported since this can be done
    directly in matplotlib with ``plt.step(...)``.

    """
    return pc_curve(*args, **kwargs)


@deprecated("This function is deprecated, use pc_curve instead")
def pc_curve_from_mio(*args, **kwargs):
    r"""
    This function is deprecated.  Use ``pc_curve`` instead.
    """
    return pc_curve(*args, **kwargs)


def pc_curve(im, sizes=None, pc=None, seq=None,
             sigma=0.072, theta=180, voxel_size=1):
    r"""
    Produces a Pc-Snwp curve given a map of meniscus radii or capillary
    pressures at which each voxel was invaded

    Parameters
    ----------
    im : ndarray
        The voxel image of the porous media with ``True`` values indicating
        the void space
    sizes : ndarray, optional
        An image containing the sphere radii at which each voxel was invaded
        during an invasion experiment.
    pc : ndarray, optional
        An image containing the capillary pressures at which each voxel was
        invaded during an invasion experiment.
    seq : ndarray, optional
        An image containing invasion sequence values, such as that returned
        from the ``ibip`` function.
    sigma : float, optional
        The surface tension of the fluid-fluid system of interest.
        This argument is ignored if ``pc`` are specified, otherwise it
        is used in the Washburn equation to convert ``sizes`` to capillary
        pc.
    theta : float
        The contact angle measured through the invading phase in degrees.
        This argument is ignored if ``pc`` are specified, otherwise it
        is used in the Washburn equation to convert ``sizes`` to capillary
        pressures.
    voxel_size : float
        The voxel resolution of the image.
        This argument is ignored if ``pc`` are specified, otherwise it
        is used in the Washburn equation to convert ``sizes`` to capillary
        pressures.

    Returns
    -------
    pc_curve : Results object
        A custom object with the following data added as named attributes:

        ==================  ===================================================
        Attribute           Description
        ==================  ===================================================
        pc                  The capillary pressure, either as given in
                            ``pc`` or computed from ``sizes`` (see
                            Notes).
        snwp                The fraction of void space filled by non-wetting
                            phase at each pressure in ``pc``
        ==================  ===================================================

    Notes
    -----
    If ``sizes`` is provided, then the Washburn equation is used to convert
    the radii to capillary pressures, using the given ``sigma`` and ``theta``
    values, along with the ``voxel_size`` if the values are in voxel radii.
    For more control over how capillary pressure model, it can be computed by
    hand, for example:

        $$ pc = \frac{-2*0.072*np.cos(np.deg2rad(180))}{sizes \cdot voxel_size} $$

    then passed in as the ``pc`` argument.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/pc_curve.html>`_
    to view online example.

    """
    tqdm = get_tqdm()
    if seq is not None:
        seqs = np.unique(seq)[1:]
        x = []
        y = []
        with tqdm(seqs, **settings.tqdm) as pbar:
            for n in seqs:
                pbar.update()
                mask = seq == n
                # The following assumes only one size found, which was confirmed
                r = sizes[mask][0]*voxel_size
                pc = -2*sigma*np.cos(np.deg2rad(theta))/r
                x.append(pc)
                snwp = ((seq <= n)*(seq > 0) *
                        (im == 1)).sum(dtype=np.int64)/im.sum(dtype=np.int64)
                y.append(snwp)
        pc_curve = Results()
        pc_curve.pc = x
        pc_curve.snwp = y
    elif sizes is not None:
        if im is None:
            im = ~(sizes == 0)
        sz = np.unique(sizes)[:0:-1]
        sz = np.hstack((sz[0]*2, sz))
        x = []
        y = []
        with tqdm(sz, **settings.tqdm) as pbar:
            for n in sz:
                pbar.update()
                r = n*voxel_size
                pc = -2*sigma*np.cos(np.deg2rad(theta))/r
                x.append(pc)
                snwp = ((sizes >= n)*(im == 1)).sum(dtype=np.int64)/im.sum(dtype=np.int64)
                y.append(snwp)
        pc_curve = Results()
        pc_curve.pc = x
        pc_curve.snwp = y
    elif pc is not None:
        Ps = np.unique(pc[im])
        # Utilize the fact that -inf and +inf will be at locations 0 & -1 in Ps
        if Ps[-1] == np.inf:
            Ps[-1] = Ps[-2]*2
        if Ps[0] == -np.inf:
            Ps[0] = Ps[1] - np.abs(Ps[1]/2)
        else:
            # Add a point at begining to ensure curve starts a 0, if no residual
            Ps = np.hstack((Ps[0] - np.abs(Ps[0]/2), Ps))
        y = []
        Vp = im.sum(dtype=np.int64)
        temp = pc[im]
        for p in tqdm(Ps, **settings.tqdm):
            y.append((temp <= p).sum(dtype=np.int64)/Vp)
        pc_curve = Results()
        pc_curve.pc = Ps
        pc_curve.snwp = y
    return pc_curve


def satn_profile(satn, s, axis=0, span=10, mode='tile'):
    r"""
    Computes a saturation profile from an image of fluid invasion

    Parameters
    ----------
    satn : ndarray
        An image with each voxel indicating the saturation upon its
        invasion.  0's are treated as solid and -1's are treated as uninvaded
        void space.
    s : scalar
        The global saturation value for which the profile is desired
    axis : int
        The axis along which to profile should be measured
    span : int
        The number of layers to include in the moving average saturation
        calculation.
    mode : str
        How the moving average should be applied. Options are:

        ======== ==============================================================
        mode     description
        ======== ==============================================================
        'tile'   The average is computed for discrete non-overlapping
                 tiles of a size given by ``span``
        'slide'  The average is computed in a moving window starting at
                 ``span/2`` and sliding by a single voxel. This method
                 provides more data points but is slower.
        ======== ==============================================================

    Returns
    -------
    results : dataclass
        Results is a custom porespy class with the following attributes:

        ============= =========================================================
        Attribute     Description
        ============= =========================================================
        position      The position along the given axis at which saturation
                      values are computed.  The units are in voxels.
        saturation    The local saturation value at each position.
        ============= =========================================================

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/satn_profile.html>`_
    to view online example.
    """
    # @numba.njit()
    def func(satn, s, axis, span, mode):
        span = max(1, span)
        satn = np.swapaxes(satn, 0, axis)
        if mode == 'tile':
            y = np.zeros(int(satn.shape[0]/span))
            z = np.zeros_like(y)
            for i in range(int(satn.shape[0]/span)):
                void = satn[i*span:(i+1)*span, ...] != 0
                nwp = (satn[i*span:(i+1)*span, ...] < s) \
                    *(satn[i*span:(i+1)*span, ...] > 0)
                y[i] = nwp.sum(dtype=np.int64)/void.sum(dtype=np.int64)
                z[i] = i*span + (span-1)/2
        if mode == 'slide':
            y = np.zeros(int(satn.shape[0]-span))
            z = np.zeros_like(y)
            for i in range(int(satn.shape[0]-span)):
                void = satn[i:i+span, ...] != 0
                nwp = (satn[i:i+span, ...] < s)*(satn[i:i+span, ...] > 0)
                y[i] = nwp.sum(dtype=np.int64)/void.sum(dtype=np.int64)
                z[i] = i + (span-1)/2
        return z, y

    z, y = func(satn=satn, s=s, axis=axis, span=span, mode=mode)

    class results(Results):
        r"""

        Attributes
        ----------
        position : ndarray
            The position along the given axis at which saturation values are
            computed.  The units are in voxels.
        saturation : ndarray
            The computed saturation value at each position

        """
        position = z
        saturation = y

    return results


def find_h(saturation, position=None, srange=[0.01, 0.99]):
    r"""
    Given a saturation profile, compute the height between given bounds

    Parameters
    ----------
    saturation : array_like
        A list of saturation values as function of ``position``
    position : array_like, optional
        A list of positions corresponding to each saturation.  If not provided
        then each value in ``saturation`` is assumed to be separated by 1 voxel.
    srange : list
        The minimum and maximum value of saturation to consider as the start
        and end of the profile

    Returns
    -------
    h : scalar
        The height of the two-phase zone

    See Also
    --------
    satn_profile

    Notes
    -----
    The ``satn_profile`` function can be used to obtain the ``saturation``
    and ``position`` from an image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/metrics/reference/find_h.html>`_
    to view online example.

    """
    r = Results()
    r.valid = True
    # First ensure saturation generally descends from left to right
    if np.mean(saturation[:10]) < np.mean(saturation[-10:]):
        saturation = np.flip(saturation, axis=0)
    # Ensure requested saturation limits actually exist
    if (min(srange) < min(saturation)) or (max(srange) > max(saturation)):
        srange = max(min(srange), min(saturation)), min(max(srange), max(saturation))
        r.valid = False
        logger.warning(f'The requested saturation range was adjusted to {srange}'
                        ' to accomodate data')
    # Find zmax
    x = saturation >= max(srange)
    zmax = np.where(x)[0][-1]
    y = saturation <= min(srange)
    zmin = np.where(y)[0][0]
    # If position array was given, index into it
    if position is not None:
        zmax = position[zmax]
        zmin = position[zmin]

    # Add remaining data to results object
    r.zmax = zmax
    r.zmin = zmin
    r.smax = max(srange)
    r.smin = min(srange)
    r.h = abs(zmax-zmin)

    return r



import numpy as np


__all__ = ['faces', 'borders']


def faces(shape, inlet=None, outlet=None):
    r"""
    Generate an image with ``True`` values on the specified ``inlet`` and
    ``outlet`` faces

    Parameters
    ----------
    shape : list
        The ``[x, y, z (optional)]`` shape to generate. This will likely
        be obtained from ``im.shape`` where ``im`` is the image for which
        an array of faces is required.
    inlet : int
        The axis where the faces should be added (e.g. ``inlet=0`` will
        put ``True`` values on the ``x=0`` face). A value of ``None``
        (default) bypasses the addition of inlets.
    outlet : int
        Same as ``inlet`` except for the outlet face. This is optional. It
        can be be applied at the same time as ``inlet``, instead of
        ``inlet`` (if ``inlet`` is set to ``None``), or ignored
        (if ``outlet = None``).

    Returns
    -------
    faces : ndarray
        A boolean image of the given ``shape`` with ``True`` values on the
        specified ``inlet`` and/or ``outlet`` face(s).

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/faces.html>`_
    to view online example.

    """
    im = np.zeros(shape, dtype=bool)
    # Parse inlet and outlet
    if inlet is not None:
        im = np.swapaxes(im, 0, inlet)
        im[0, ...] = True
        im = np.swapaxes(im, 0, inlet)
    if outlet is not None:
        im = np.swapaxes(im, 0, outlet)
        im[-1, ...] = True
        im = np.swapaxes(im, 0, outlet)
    if (inlet is None) and (outlet is None):
        raise Exception('Both inlet and outlet were given as None')
    return im


def borders(shape, thickness=1, mode='edges'):
    r"""
    Creates an array of specified size with corners, edges or faces
    labelled as ``True``.

    This can be used as mask to manipulate values laying on the perimeter
    of an image.

    Parameters
    ----------
    shape : array_like
        The shape of the array to return.  Can be either 2D or 3D.
    thickness : scalar (default is 1)
        The number of pixels/voxels layers to place along perimeter.
    mode : string
        The type of border to create.  Options are 'faces', 'edges'
        (default) and 'corners'.  In 2D 'corners' and 'edges' give the
        same result.

    Returns
    -------
    image : ndarray
        An ndarray of specified shape with ``True`` values at the
        perimeter and ``False`` elsewhere

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/borders.html>`_
    to view online example.

    """
    ndims = len(shape)
    t = thickness
    border = np.ones(shape, dtype=bool)
    if mode == 'faces':
        if ndims == 2:
            border[t:-t, t:-t] = False
        if ndims == 3:
            border[t:-t, t:-t, t:-t] = False
    elif mode == 'edges':
        if ndims == 2:
            border[t:-t, 0::] = False
            border[0::, t:-t] = False
        if ndims == 3:
            border[0::, t:-t, t:-t] = False
            border[t:-t, 0::, t:-t] = False
            border[t:-t, t:-t, 0::] = False
    elif mode == 'corners':
        if ndims == 2:
            border[t:-t, 0::] = False
            border[0::, t:-t] = False
        if ndims == 3:
            border[t:-t, 0::, 0::] = False
            border[0::, t:-t, 0::] = False
            border[0::, 0::, t:-t] = False
    return border



r"""

Collection of functions for extracting quantitative information from images
###########################################################################

This submodule contains functions for determining key metrics about an
image. Typically these are applied to an image after applying a filter,
but a few functions can be applied directly to the binary image.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    metrics.boxcount
    metrics.chord_counts
    metrics.chord_length_distribution
    metrics.find_h
    metrics.lineal_path_distribution
    metrics.mesh_surface_area
    metrics.mesh_volume
    metrics.pc_curve
    metrics.pc_curve_from_ibip
    metrics.pc_curve_from_mio
    metrics.phase_fraction
    metrics.pore_size_distribution
    metrics.porosity
    metrics.porosity_profile
    metrics.prop_to_image
    metrics.props_to_DataFrame
    metrics.radial_density_distribution
    metrics.region_interface_areas
    metrics.region_surface_areas
    metrics.region_volumes
    metrics.regionprops_3D
    metrics.representative_elementary_volume
    metrics.satn_profile
    metrics.two_point_correlation

"""

from ._regionprops import *
from ._funcs import *
from ._meshtools import *



import dask.array as da
import inspect as insp
import logging
import numpy as np
from numba import njit, prange
from edt import edt
import scipy.ndimage as spim
import scipy.spatial as sptl
from skimage.segmentation import watershed
from skimage.morphology import ball, disk, square, cube
from porespy.tools import _check_for_singleton_axes
from porespy.tools import extend_slice, ps_rect, ps_round
from porespy.tools import Results
from porespy.tools import get_tqdm
from porespy.filters import chunked_func
from porespy import settings


__all__ = [
    "snow_partitioning",
    "snow_partitioning_n",
    "snow_partitioning_parallel",
    "find_peaks",
    "reduce_peaks",
    "trim_nearby_peaks",
    "trim_saddle_points",
    "trim_saddle_points_legacy",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


def snow_partitioning(im, dt=None, r_max=4, sigma=0.4, peaks=None):
    r"""
    Partition the void space into pore regions using a marker-based
    watershed algorithm

    Parameters
    ----------
    im : array_like
        A boolean image of the domain, with ``True`` indicating the pore
        space and ``False`` elsewhere.
    dt : array_like, optional
        The distance transform of the pore space.  This is done
        automatically if not provided, but if the distance transform has
        already been computed then supplying it can save some time.
    r_max : int
        The radius of the spherical structuring element to use in the
        Maximum filter stage that is used to find peaks. The default is 4.
    sigma : float
        The standard deviation of the Gaussian filter used in step 1. The
        default is 0.4.  If 0 is given then the filter is not applied,
        which is useful if a distance transform is supplied as the ``im``
        argument that has already been processed.
    peaks : ndarray, optional
        Optionally, it is possible to supply an array containing peaks, which
        are used as markers in the watershed segmentation. If a boolean array
        is received (``True`` indicating peaks), then ``scipy.ndimage.label``
        with cubic connectivity is used to label them. If an integer array is
        received then it is assumed the peaks have already been labelled.
        This allows for comparison of peak finding algorithms for instance.
        If this argument is provided, then ``r_max`` and ``sigma`` are ignored
        since these are specfically used in the peak finding process.

    Returns
    -------
    results : Results object
        A custom object with the following data as attributes:

        ============ ==========================================================
        Item         Description
        ============ ==========================================================
        ``im``       The binary image of the void space
        ``dt``       The distance transform of the image
        ``peaks``    The peaks of the distance transform after applying the
                     steps of the SNOW algorithm
        ``regions``  The void space partitioned into pores using a marker
                     based watershed with the peaks found by the SNOW algorithm
        ============ ==========================================================

    Notes
    -----
    The SNOW network extraction algorithm (Sub-Network of an
    Over-segmented Watershed) was designed to handle to perculiarities of
    high porosity materials, but it applies well to other materials as
    well.

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction
    algorithm using marker-based watershed segmenation".  Physical Review
    E. (2017)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/snow_partitioning.html>`_
    to view online example.

    """
    logger.info("Beginning SNOW algorithm")
    im_shape = np.array(im.shape)
    if im.dtype is not bool:
        logger.info("Converting supplied image to boolean")
        im = im > 0
    if dt is None:
        logger.info("Peforming distance transform")
        if np.any(im_shape == 1):
            dt = edt(im.squeeze())
            dt = dt.reshape(im_shape)
        else:
            dt = edt(im)

    if peaks is None:
        if sigma > 0:
            logger.info(f"Applying Gaussian blur with sigma = {sigma}")
            dt_blurred = spim.gaussian_filter(input=dt, sigma=sigma)*im
        else:
            dt_blurred = np.copy(dt)
        peaks = find_peaks(dt=dt_blurred, r_max=r_max)

        logger.debug(f"Initial number of peaks: {spim.label(peaks)[1]}")
        peaks = trim_saddle_points(peaks=peaks, dt=dt)
        logger.debug(f"Peaks after trimming saddle points: {spim.label(peaks)[1]}")
        peaks = trim_nearby_peaks(peaks=peaks, dt=dt)
        logger.debug(f"Peaks after trimming nearby points: {spim.label(peaks)[1]}")
    peaks, N = spim.label(peaks > 0, structure=ps_rect(3, im.ndim))
    regions = watershed(image=-dt, markers=peaks)
    tup = Results()
    tup.im = im
    tup.dt = dt
    tup.peaks = peaks
    tup.regions = regions * (im > 0)
    return tup


def snow_partitioning_n(im, r_max=4, sigma=0.4, peaks=None):
    r"""
    This function partitions an imaging oontain an arbitrary number of
    phases into regions using a marker-based watershed segmentation.

    Parameters
    ----------
    im : ndarray
        Image of porous material where each phase is represented by unique
        integer starting from 1 (0's are ignored).
    r_max : scalar
        The radius of the spherical structuring element to use in the
        Maximum filter stage that is used to find peaks. The default is 4.
    sigma : scalar
        The standard deviation of the Gaussian filter used. The default is
        0.4. If 0 is given then the filter is not applied.
    peaks : ndarray, optional
        Optionally, it is possible to supply an array containing peaks, which
        are used as markers in the watershed segmentation. Must be a boolean
        array with ``True`` indicating peaks; ``scipy.ndimage.label``
        with cubic connectivity is used to label them. If this argument is
        provided then ``r_max`` and ``sigma`` are ignored, since these are
        specfically used in the peak finding process.

    Returns
    -------
    results : Results object
        A custom object with the following data as attributes:

        - 'im'
            The original image of the porous material

        - 'dt'
            The combined distance transform in alll phases of the image

        - 'phase_max_label'
            The list of max label of each phase in order to
            distinguish between each other

        - 'regions'
            The partitioned regions of n phases using a marker
            based watershed with the peaks found by the SNOW algorithm

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction
    algorithm using marker-based watershed segmentation".  Physical Review
    E. (2017)

    [2] Khan, ZA et al. "Dual network extraction algorithm to investigate
    multiple transport processes in porous materials: Image-based modeling
    of pore and grain-scale processes". Computers in Chemical Engineering.
    (2019)

    See Also
    --------
    snow_partitioning

    Notes
    -----
    In principle it is possible to perform a distance transform on each
    phase separately, merge these into a single image, then apply the
    watershed only once. This, however, has been found to create edge
    artifacts between regions arising from the way watershed handles
    plateaus in the distance transform. To overcome this, this function
    applies the watershed to each of the distance transforms separately,
    then merges the segmented regions back into a single image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/snow_partitioning_n.html>`_
    to view online example.

    """
    # Perform snow on each phase and merge all segmentation and dt together
    phases_num = np.unique(im).astype(int)
    phases_num = phases_num[phases_num > 0]
    combined_dt = 0
    combined_region = 0
    _peaks = np.zeros_like(im, dtype=int)
    num = [0]
    for i, j in enumerate(phases_num):
        logger.info(f"Processing Phase {j}")
        # Isolate active phase from image
        phase = im == j
        # Limit peaks to active phase only
        temp = peaks*phase if peaks is not None else None
        phase_snow = snow_partitioning(phase, dt=None, r_max=r_max,
                                       sigma=sigma, peaks=temp)
        combined_dt += phase_snow.dt
        phase_snow.regions *= phase_snow.im
        phase_snow.regions += num[i]
        phase_ws = phase_snow.regions * phase_snow.im
        phase_ws[phase_ws == num[i]] = 0
        combined_region += phase_ws
        _peaks = _peaks + phase_snow.peaks + (phase_snow.peaks > 0)*num[i]
        num.append(np.amax(combined_region))

    tup = Results()
    tup.im = im
    tup.dt = combined_dt
    tup.phase_max_label = num[1:]
    tup.regions = combined_region
    tup.peaks = _peaks
    return tup


def find_peaks(dt, r_max=4, strel=None, sigma=None, divs=1):
    r"""
    Finds local maxima in the distance transform

    Parameters
    ----------
    dt : ndarray
        The distance transform of the pore space.  This may be calculated
        and filtered using any means desired.
    r_max : scalar
        The radius of the spherical element used in the maximum filter.
        This controls the localness of any maxima. The default is 4 voxels.
    strel : ndarray
        Instead of supplying ``r_max``, this argument allows a custom
        structuring element allowing control over both size and shape.
    sigma : float or list of floats
        If given, then a gaussian filter is applied to the distance transform
        using this value for the kernel
        (i.e. ``scipy.ndimage.gaussian_filter(dt, sigma)``)
    divs : int or array_like
        The number of times to divide the image for parallel processing.
        If ``1`` then parallel processing does not occur.  ``2`` is
        equivalent to ``[2, 2, 2]`` for a 3D image. The number of cores
        used is specified in ``porespy.settings.ncores`` and defaults to
        all cores.

    Returns
    -------
    image : ndarray
        An array of booleans with ``True`` values at the location of any
        local maxima.

    Notes
    -----
    It is also possible ot the ``peak_local_max`` function from the
    ``skimage.feature`` module as follows:

    ``peaks = peak_local_max(image=dt, min_distance=r, exclude_border=0,
    indices=False)``

    The *skimage* function automatically uses a square structuring element
    which is significantly faster than using a circular or spherical
    element.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/find_peaks.html>`_
    to view online example.

    """
    im = dt > 0
    _check_for_singleton_axes(im)
    if strel is None:
        strel = ps_round(r=r_max, ndim=im.ndim)
    if sigma is not None:
        dt = spim.gaussian_filter(dt, sigma=sigma)
    parallel = False
    if isinstance(divs, int):
        divs = [divs]*len(im.shape)
    if np.any(np.array(divs) > 1):
        parallel = True
        logger.info(f'Performing {insp.currentframe().f_code.co_name} in parallel')
    if parallel:
        overlap = max(strel.shape)
        mx = chunked_func(func=spim.maximum_filter, overlap=overlap,
                          im_arg='input', input=dt + 2.0 * (~im),
                          footprint=strel,
                          cores=settings.ncores, divs=divs)
    else:
        # The "2 * (~im)" sets solid voxels to 2 so peaks are not found
        # at the void/solid interface
        mx = spim.maximum_filter(dt + 2.0 * (~im), footprint=strel)
    peaks = (dt == mx) * im
    return peaks


def reduce_peaks(peaks):
    r"""
    Any peaks that are broad or elongated are replaced with a single voxel
    that is located at the center of mass of the original voxels.

    Parameters
    ----------
    peaks : ndarray
        An image containing ``True`` values indicating peaks in the
        distance transform

    Returns
    -------
    image : ndarray
        An array with the same number of isolated peaks as the original
        image, but fewer total ``True`` voxels.

    Notes
    -----
    The center of mass of a group of voxels is used as the new single
    voxel, so if the group has an odd shape (like a horse shoe), the new
    voxel may *not* lie on top of the original set.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/reduce_peaks.html>`_
    to view online example.

    """
    if peaks.ndim == 2:
        strel = square
    else:
        strel = cube
    markers, N = spim.label(input=peaks, structure=strel(3))
    inds = spim.measurements.center_of_mass(
        input=peaks, labels=markers, index=np.arange(1, N + 1)
    )
    inds = np.floor(inds).astype(int)
    # Centroid may not be on old pixel, so create a new peaks image
    peaks_new = np.zeros_like(peaks, dtype=bool)
    peaks_new[tuple(inds.T)] = True
    return peaks_new


def trim_saddle_points(peaks, dt, maxiter=20):
    r"""
    Removes peaks that were mistakenly identified because they lied on a
    saddle or ridge in the distance transform that was not actually a true
    local peak.

    Parameters
    ----------
    peaks : ndarray
        A boolean image containing ``True`` values to mark peaks in the
        distance transform (``dt``)
    dt : ndarray
        The distance transform of the pore space for which the peaks
        are sought.
    maxiter : int
        The number of iteration to use when finding saddle points.
        The default value is 20.

    Returns
    -------
    image : ndarray
        An image with fewer peaks than the input image

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction algorithm
    using marker-based watershed segmentation".  Physical Review E. (2017)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/trim_saddle_points.html>`_
    to view online example.

    """
    new_peaks = np.zeros_like(peaks, dtype=bool)
    if dt.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube
    labels, N = spim.label(peaks > 0)
    slices = spim.find_objects(labels)
    for i, s in tqdm(enumerate(slices), **settings.tqdm):
        sx = extend_slice(s, shape=peaks.shape, pad=maxiter)
        peaks_i = labels[sx] == i + 1
        dt_i = dt[sx]
        im_i = dt_i > 0
        iters = 0
        while iters < maxiter:
            iters += 1
            peaks_dil = spim.binary_dilation(input=peaks_i, structure=cube(3))
            peaks_max = peaks_dil * np.amax(dt_i * peaks_dil)
            peaks_extended = (peaks_max == dt_i) * im_i
            if np.all(peaks_extended == peaks_i):
                new_peaks[sx] += peaks_i
                break  # Found a true peak
            elif np.sum(peaks_extended * peaks_i, dtype=np.int64) == 0:
                break  # Found a saddle point
            peaks_i = peaks_extended
        if iters >= maxiter:
            logger.debug(
                "Maximum number of iterations reached, consider "
                + "running again with a larger value of max_iters"
            )
    return new_peaks*peaks


def trim_saddle_points_legacy(peaks, dt, maxiter=10):
    r"""
    Removes peaks that were mistakenly identified because they lied on a
    saddle or ridge in the distance transform that was not actually a true
    local peak.

    Parameters
    ----------
    peaks : ND-array
        A boolean image containing True values to mark peaks in the distance
        transform (``dt``)
    dt : ND-array
        The distance transform of the pore space for which the true peaks are
        sought.
    maxiter : int
        The maximum number of iterations to run while eroding the saddle
        points.  The default is 10, which is usually not reached; however,
        a warning is issued if the loop ends prior to removing all saddle
        points.

    Returns
    -------
    image : ND-array
        An image with fewer peaks than the input image

    Notes
    -----
    This version of the function was included in versions of PoreSpy < 2. It
    is too aggressive in trimming peaks, so was rewritten for PoreSpy >= 2.
    This is included here for legacy reasons

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction algorithm
    using marker-based watershed segmenation".  Physical Review E. (2017)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/trim_saddle_points_legacy.html>`_
    to view online example.
    """
    new_peaks = np.zeros_like(peaks, dtype=bool)
    if dt.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube
    labels, N = spim.label(peaks > 0)
    slices = spim.find_objects(labels)
    for i, s in tqdm(enumerate(slices), **settings.tqdm):
        sx = extend_slice(s, shape=peaks.shape, pad=10)
        peaks_i = labels[sx] == i + 1
        dt_i = dt[sx]
        im_i = dt_i > 0
        iters = 0
        while iters < maxiter:
            iters += 1
            peaks_dil = spim.binary_dilation(input=peaks_i, structure=cube(3))
            peaks_max = peaks_dil * np.amax(dt_i * peaks_dil)
            peaks_extended = (peaks_max == dt_i) * im_i
            if np.all(peaks_extended == peaks_i):
                break  # Found a true peak
            elif np.sum(peaks_extended * peaks_i, dtype=np.int64) == 0:
                peaks_i = False
                break  # Found a saddle point
            # The following line was also missing from the original.  It has
            # no effect on the result but without this the "maxiters" is
            # reached very often.
            peaks_i = peaks_extended
        # The following line is essentially a bug.  It should be:
        # peaks[s] += peaks_i. Without the += the peaks_i image overwrites
        # the entire slice s, which may include other peaks that are within
        # 10 voxels due to the padding of s with extend_slice.
        new_peaks[sx] = peaks_i
        if iters >= maxiter:
            logger.debug(
                "Maximum number of iterations reached, consider "
                + "running again with a larger value of max_iters"
            )
    return new_peaks*peaks


def trim_nearby_peaks(peaks, dt, f=1):
    r"""
    Removes peaks that are nearer to another peak than to solid

    Parameters
    ----------
    peaks : ndarray
        A image containing nonzeros values indicating peaks in the distance
        transform (``dt``).  If ``peaks`` is boolean, a boolean is returned;
        if ``peaks`` have already been labelled, then the original labels
        are returned, missing the trimmed peaks.
    dt : ndarray
        The distance transform of the pore space
    f : scalar
        Controls how close peaks must be before they are considered near
        to each other. Sets of peaks are tagged as too near if
        ``d_neighbor < f * d_solid``.

    Returns
    -------
    image : ndarray
        An array the same size and type as ``peaks`` containing a subset of
        the peaks in the original image.

    Notes
    -----
    Each pair of peaks is considered simultaneously, so for a triplet of nearby
    peaks, each pair is considered.  This ensures that only the single peak
    that is furthest from the solid is kept.  No iteration is required.

    References
    ----------
    [1] Gostick, J. "A versatile and efficient network extraction
    algorithm using marker-based watershed segmenation". Physical Review
    E. (2017)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/trim_nearby_peaks.html>`_
    to view online example.

    """
    if dt.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube

    labels, N = spim.label(peaks > 0, structure=cube(3))
    crds = spim.measurements.center_of_mass(peaks > 0, labels=labels,
                                            index=np.arange(1, N + 1))
    crds = np.vstack(crds).astype(int)  # Convert to numpy array of ints
    L = dt[tuple(crds.T)]  # Get distance to solid for each peak
    # Add tiny amount to joggle points to avoid equal distances to solid
    # arange was added instead of random values so the results are repeatable
    L = L + np.arange(len(L))*1e-6

    tree = sptl.KDTree(data=crds)
    # Find list of nearest peak to each peak
    temp = tree.query(x=crds, k=2)
    nearest_neighbor = temp[1][:, 1]
    dist_to_neighbor = temp[0][:, 1]
    del temp, tree  # Free-up memory
    hits = np.where(dist_to_neighbor <= f * L)[0]
    # Drop peak that is closer to the solid than it's neighbor
    drop_peaks = []
    for i in hits:
        if L[i] < L[nearest_neighbor[i]]:
            drop_peaks.append(i)
        else:
            drop_peaks.append(nearest_neighbor[i])
    drop_peaks = np.unique(drop_peaks)

    new_peaks = ~np.isin(labels, drop_peaks+1)*peaks
    return new_peaks


def _estimate_overlap(im, mode='dt', zoom=0.25):
    logger.info('Calculating overlap thickness')
    if mode == 'watershed':
        rev = spim.interpolation.zoom(im, zoom=zoom, order=0)
        rev = rev > 0
        dt = edt(rev, parallel=0)
        rev_snow = snow_partitioning(rev, dt=dt)
        labels, counts = np.unique(rev_snow, return_counts=True)
        node = np.where(counts == counts[1:].max())[0][0]
        slices = spim.find_objects(rev_snow)
        overlap = max(rev_snow[slices[node - 1]].shape) / (zoom * 2.0)
    if mode == 'dt':
        dt = edt((im > 0), parallel=0)
        overlap = dt.max()
    return overlap


def snow_partitioning_parallel(im,
                               r_max=4,
                               sigma=0.4,
                               divs=2,
                               overlap=None,
                               cores=None,
                               ):
    r"""
    Performs SNOW algorithm in parallel (or serial) to reduce time
    (or memory usage) by geomertirc domain decomposition of large images.

    Parameters
    ----------
    im : ndarray
        A binary image of porous media with 'True' values indicating
        phase of interest.
    overlap : float (optional)
        The amount of overlap to apply between chunks.  If not provided it
        will be estiamted using ``porespy.tools.estimate_overlap`` with
        ``mode='dt'``.
    divs : list or int
        Number of domains each axis will be divided. Options are:
          - scalar: it will be assigned to all axis.
          - list: each respective axis will be divided by its
            corresponding number in the list. For example [2, 3, 4] will
            divide z, y and x axis to 2, 3, and 4 respectively.
    cores : int or None
        Number of cores that will be used to parallel process all domains.
        If ``None`` then all cores will be used but user can specify any
        integer values to control the memory usage.  Setting value to 1
        will effectively process the chunks in serial to minimize memory
        usage.

    Returns
    -------
    regions : ndarray
        Partitioned image of segmentated regions with unique labels. Each
        region correspond to pore body while intersection with other
        region correspond throat area.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/snow_partitioning_parallel.html>`_
    to view online example.

    """
    # Adjust image shape according to specified dimension
    if isinstance(divs, int):
        divs = [divs for i in range(im.ndim)]
    shape = []
    for i in range(im.ndim):
        shape.append(divs[i] * (im.shape[i] // divs[i]))

    if tuple(shape) != im.shape:
        for i in range(im.ndim):
            im = im.swapaxes(0, i)
            im = im[:shape[i], ...]
            im = im.swapaxes(i, 0)
        logger.debug(f'Image was cropped to shape {shape}')

    # Get overlap thickness from distance transform
    chunk_shape = (np.array(shape) / np.array(divs)).astype(int)
    logger.info('Beginning parallel SNOW algorithm...')

    if overlap is None:
        overlap = _estimate_overlap(im, mode='dt')
    overlap = overlap / 2.0
    logger.debug(f'Overlap thickness: {int(2 * overlap)} voxels')

    dt = edt((im > 0), parallel=0)

    # Get overlap and trim depth of all image dimension
    depth = {}
    trim_depth = {}
    for i in range(im.ndim):
        depth[i] = int(2.0 * overlap)
        trim_depth[i] = int(2.0 * overlap) - 1

    # Applying SNOW to image chunks
    regions = da.from_array(dt, chunks=chunk_shape)
    regions = da.overlap.overlap(regions, depth=depth, boundary='none')
    regions = regions.map_blocks(_snow_chunked, r_max=r_max,
                                 sigma=sigma, dtype=dt.dtype)
    regions = da.overlap.trim_internal(regions, trim_depth, boundary='none')
    # TODO: use dask ProgressBar once compatible w/ logging.
    logger.info('Applying snow to image chunks')
    regions = regions.compute(num_workers=cores)

    # Relabelling watershed chunks
    logger.info('Relabelling watershed chunks')
    regions = relabel_chunks(im=regions, chunk_shape=chunk_shape)

    # Stitching watershed chunks
    logger.info('Stitching watershed chunks')
    regions = _watershed_stitching(im=regions, chunk_shape=chunk_shape)
    tup = Results()
    tup.im = im
    tup.dt = dt
    tup.regions = regions
    return tup


def _pad(im, pad_width=1, constant_value=0):
    r"""
    Pad the image with a constant values and width.

    Parameters
    ----------
    im : ndarray
        The image that requires padding
    pad_width : int
        The number of values that will be padded from the edges. Default
        values is 1.
    contant_value : int
        Pads with the specified constant value

    Returns
    -------
    output: ndarray
        Padded image with same dimnesions as provided image

    """
    shape = np.array(im.shape)
    pad_shape = shape + (2 * pad_width)
    temp = np.zeros(pad_shape, dtype=np.uint32)
    if constant_value != 0:
        temp = temp + constant_value
    if im.ndim == 3:
        temp[pad_width: -pad_width,
             pad_width: -pad_width,
             pad_width: -pad_width] = im
    elif im.ndim == 2:
        temp[pad_width: -pad_width,
             pad_width: -pad_width] = im
    else:
        temp[pad_width: -pad_width] = im

    return temp


def relabel_chunks(im, chunk_shape):
    r"""
    Assign new labels to each chunk or sub-domain of actual image. This
    prevents from two or more regions to have same label.

    Parameters
    ----------
    im: ndarray
        Actual image that contains repeating labels in chunks/sub-domains.
    chunk_shape: tuple
        The shape of chunk that will be relabeled in actual image. Note
        the chunk shape should be a multiple of actual image shape
        otherwise some labels will not be relabeled.

    Returns
    -------
    output : ndarray
        Relabeled image with unique label assigned to each region.

    """
    im = _pad(im, pad_width=1)
    im_shape = np.array(im.shape, dtype=np.uint32)
    max_num = 0
    c = np.array(chunk_shape, dtype=np.uint32) + 2
    num = (im_shape / c).astype(int)

    if im.ndim == 3:
        for z in range(num[0]):
            for y in range(num[1]):
                for x in range(num[2]):
                    chunk = im[z * c[0]: (z + 1) * c[0],
                               y * c[1]: (y + 1) * c[1],
                               x * c[2]: (x + 1) * c[2]]
                    chunk += max_num
                    chunk[chunk == max_num] = 0
                    max_num = chunk.max()
                    im[z * c[0]: (z + 1) * c[0],
                       y * c[1]: (y + 1) * c[1],
                       x * c[2]: (x + 1) * c[2]] = chunk
    else:
        for y in range(num[0]):
            for x in range(num[1]):
                chunk = im[y * c[0]: (y + 1) * c[0],
                           x * c[1]: (x + 1) * c[1]]
                chunk += max_num
                chunk[chunk == max_num] = 0
                max_num = chunk.max()
                im[y * c[0]: (y + 1) * c[0],
                   x * c[1]: (x + 1) * c[1]] = chunk

    return im


def _trim_internal_slice(im, chunk_shape):
    r"""
    Delete extra slices from image that were used to stitch two or more
    chunks together.

    Parameters:
    -----------
    im :  ndarray
        image that contains extra slices in x, y, z direction.
    chunk_shape : tuple
        The shape of the chunk from which image is subdivided.

    Return:
    -------
    output : ndarray
        Image without extra internal slices. The shape of the image will
        be same as input image provided for waterhsed segmentation.

    """
    im_shape = np.array(im.shape, dtype=np.uint32)
    c1 = np.array(chunk_shape, dtype=np.uint32) + 2
    c2 = np.array(chunk_shape, dtype=np.uint32)
    num = (im_shape / c1).astype(int)
    out_shape = num * c2
    out = np.empty((out_shape), dtype=np.uint32)

    if im.ndim == 3:
        for z in range(num[0]):
            for y in range(num[1]):
                for x in range(num[2]):
                    chunk = im[z * c1[0]: (z + 1) * c1[0],
                               y * c1[1]: (y + 1) * c1[1],
                               x * c1[2]: (x + 1) * c1[2]]

                    out[z * c2[0]: (z + 1) * c2[0],
                        y * c2[1]: (y + 1) * c2[1],
                        x * c2[2]: (x + 1) * c2[2]] = chunk[1:-1, 1:-1, 1:-1]
    else:
        for y in range(num[0]):
            for x in range(num[1]):
                chunk = im[y * c1[0]: (y + 1) * c1[0],
                           x * c1[1]: (x + 1) * c1[1]]

                out[y * c2[0]: (y + 1) * c2[0],
                    x * c2[1]: (x + 1) * c2[1]] = chunk[1:-1, 1:-1]

    return out


def _watershed_stitching(im, chunk_shape):
    r"""
    Stitch individual sub-domains of watershed segmentation into one big
    segmentation with all boundary labels of each sub-domain relabeled to
    merge boundary regions.

    Parameters:
    -----------
    im : ndarray
        A worked image with watershed segmentation performed on all
        sub-domains individually.
    chunk_shape: tuple
        The shape of the sub-domain in which segmentation is performed.

    Returns
    -------
    output : ndarray
        Stitched watershed segmentation with all sub-domains merged to
        form a single watershed segmentation.

    """
    c_shape = np.array(chunk_shape)
    cuts_num = (np.array(im.shape) / c_shape).astype(np.uint32)

    for axis, num in enumerate(cuts_num):
        keys = []
        values = []
        if num > 1:
            im = im.swapaxes(0, axis)
            for i in range(1, num):
                sl = i * (chunk_shape[axis] + 3) - (i - 1)
                sl1 = im[sl - 3, ...]
                sl1_mask = sl1 > 0
                sl2 = im[sl - 1, ...] * sl1_mask
                sl1_labels = sl1.flatten()[sl1.flatten() > 0]
                sl2_labels = sl2.flatten()[sl2.flatten() > 0]
                if sl1_labels.size != sl2_labels.size:
                    raise Exception('The selected overlapping thickness is not '
                                    'suitable for input image. Change '
                                    'overlapping criteria '
                                    'or manually input value.')
                keys.append(sl1_labels)
                values.append(sl2_labels)
            im = _replace_labels(array=im, keys=keys, values=values)
            im = im.swapaxes(axis, 0)
    im = _trim_internal_slice(im=im, chunk_shape=chunk_shape)
    im = _resequence_labels(array=im)

    return im


@njit(parallel=True)
def _copy(im, output):
    r"""
    The function copy the input array and make output array that is
    allocated in different memory space. This a numba version of copy
    function of numpy. Because each element is copied using parallel
    approach this implementation is faster than numpy version of copy.

    Parameters
    ----------
    array: ndarray
        Array that needs to be copied.

    Returns
    -------
    output: ndarray
        Copied array.

    """

    if im.ndim == 3:
        for i in prange(im.shape[0]):
            for j in prange(im.shape[1]):
                for k in prange(im.shape[2]):
                    output[i, j, k] = im[i, j, k]
    elif im.ndim == 2:
        for i in prange(im.shape[0]):
            for j in prange(im.shape[1]):
                output[i, j] = im[i, j]
    else:
        for i in prange(im.shape[0]):
            output[i] = im[i]

    return output


@njit(parallel=True)
def _replace(array, keys, values, ind_sort):
    r"""
    This function replace keys elements in input array with new value
    elements. This function is used as internal function of
    replace_relabels.

    Parameters
    ----------
    array : ndarray
        Array which requires replacing labels.
    keys :  array_like
        1d array containing unique labels that need to be replaced.
    values : array_like
        1d array containing unique values that will be assigned to labels.

    Returns
    -------
    array : ndarray
        Array with replaced labels.

    """
    # ind_sort = np.argsort(keys)
    keys_sorted = keys[ind_sort]
    values_sorted = values[ind_sort]
    s_keys = set(keys)

    for i in prange(array.shape[0]):
        if array[i] in s_keys:
            ind = np.searchsorted(keys_sorted, array[i])
            array[i] = values_sorted[ind]


def _replace_labels(array, keys, values):
    r"""
    Replace labels in array provided as keys to values.

    Parameter:
    ----------
    array : ndarray
        Array which requires replacing labels
    keys :  1D-array
        The unique labels that need to be replaced
    values : 1D-array
        The unique values that will be assigned to labels

    return:
    -------
    array : ndarray
        Array with replaced labels.
    """
    a_shape = array.shape
    array = array.flatten()
    keys = np.concatenate(keys, axis=0)
    values = np.concatenate(values, axis=0)
    ind_sort = np.argsort(keys)
    _replace(array, keys, values, ind_sort)

    return array.reshape(a_shape)


@njit()
def _sequence(array, count):
    r"""
    Internal function of resequnce_labels method. This function resquence
    array elements in an ascending order using numba technique which is
    many folds faster than make contigious funcition.

    Parameters
    ----------
    array: ndarray
        1d array that needs resquencing.
    count: array_like
        1d array of zeros having same size as array.

    Returns
    -------
    array: 1d-array
        The input array with elements resequenced in ascending order

    Notes
    -----
    The output of this function is not same as make_contigous or
    relabel_sequential function of scikit-image. This function resequence
    and randomize the regions while other methods only do resequencing and
    output sorted array.

    """
    a = 1
    i = 0
    while i < (len(array)):
        data = array[i]
        if data != 0:
            if count[data] == 0:
                count[data] = a
                a += 1
        array[i] = count[data]
        i += 1


@njit(parallel=True)
def _amax(array):
    r"""
    Find largest element in an array using fast parallel numba technique.

    Parameter:
    ----------
    array: ndarray
        Array in which largest elements needs to be calcuted.

    Returns
    -------
    scalar: float or int
        The largest element value in the input array.

    """
    return np.max(array)


def _resequence_labels(array):
    r"""
    Resequence the lablels to make them contigious.

    Parameters
    ----------
    array: ndarray
        Array that requires resequencing.

    Returns
    -------
    array : ndarray
        Resequenced array with same shape as input array.

    """
    a_shape = array.shape
    array = array.ravel()
    max_num = _amax(array) + 1
    count = np.zeros(max_num, dtype=np.uint32)
    _sequence(array, count)

    return array.reshape(a_shape)


def _snow_chunked(dt, r_max=5, sigma=0.4):
    r"""
    This private version of snow is called during snow_parallel.
    """
    dt2 = spim.gaussian_filter(input=dt, sigma=sigma)
    peaks = find_peaks(dt=dt2, r_max=r_max)
    peaks = trim_saddle_points(peaks=peaks, dt=dt)
    peaks = trim_nearby_peaks(peaks=peaks, dt=dt)
    peaks, N = spim.label(peaks > 0)
    regions = watershed(image=-dt, markers=peaks)
    return regions * (dt > 0)



r"""
#######
PoreSpy
#######

PoreSpy is a package for performing image analysis on volumetric images of
porous materials.

PoreSpy consists of several key modules. Each module is consisted of
several functions. Here, you'll find a comprehensive documentation of the
modules, occasionally with basic embedded examples on how to use them.

"""

from .tools._utils import Settings as _Settings

settings = _Settings()

from . import tools
from . import filters
from . import metrics
from . import networks
from . import generators
from . import simulations
from . import visualization
from . import io
# The dns module will be deprecated in V3, in favor of simulations
from . import dns

from .visualization import imshow

import numpy as _np
_np.seterr(divide='ignore', invalid='ignore')

__version__ = tools._get_version()


def _setup_logger_rich():
    import logging
    from rich.logging import RichHandler

    FORMAT = "%(message)s"
    logging.basicConfig(
        format=FORMAT, datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True)]
    )


_setup_logger_rich()



import logging
import numpy as np
from porespy.networks import regions_to_network
from porespy.networks import add_boundary_regions
from porespy.networks import label_phases, label_boundaries
from porespy.filters import snow_partitioning, snow_partitioning_parallel
from porespy.tools import Results


__all__ = [
    "snow2",
    "_parse_pad_width",
]


logger = logging.getLogger(__name__)


def snow2(phases,
          phase_alias=None,
          boundary_width=3,
          accuracy='standard',
          voxel_size=1,
          sigma=0.4,
          r_max=4,
          peaks=None,
          parallelization={},):
    r"""
    Applies the SNOW algorithm to each phase indicated in ``phases``.

    This function is a combination of ``snow`` [1]_, ``snow_dual`` [2]_,
    ``snow_n`` [3]_, and ``snow_parallel`` [4]_ from previous versions.

    Parameters
    ----------
    phases : ndarray
        An image indicating the phase(s) of interest. A watershed is
        produced for each integer value in ``phases`` (except 0's). These
        are then combined into a single image and one network is extracted
        using ``regions_to_network``.
    phase_alias : dict
        A mapping between integer values in ``phases`` and phase name
        used to add labels to the network. For instance, asssuming a
        two-phase image, ``{1: 'void', 2: 'solid'}`` will result in the
        labels ``'pore.void'`` and ``'pore.solid'``, as well as
        ``'throat.solid_void'``, ``'throat.solid_solid'``, and
        ``'throat.void_void'``. If not provided, aliases are assumed to be
        ``{1: 'phase1', 2: 'phase2, ...}``.  Phase labels can also be
        applied afterward using ``label_phases``.
    boundary_width : depends
        Number of voxels to add to the beginning and end of each axis.
        This argument can either be a scalar or a list. If a scalar is
        passed, it will be applied to the beginning and end of all axes.
        If a list, you can specify the number of voxels for each axis
        individually. Here are some examples:

            - [0, 3, 0]: 3 voxels only applied to the y-axis.

            - [0, [0, 3], 0]: 3 voxels only applied to the end of y-axis.

            - [0, [3, 0], 0]: 3 voxels only applied to the beginning of y-axis.

        The default is to add 3 voxels on both ends of all axes. For each
        boundary width that is not 0, a label will automatically be
        applied indicating which end of which axis (i.e. ``'xmin'`` and
        ``'xmax'``).
    accuracy : string
        Controls how accurately certain properties are calculated during
        the analysis of regions in the ``regions_to_network`` function.
        Options are:

            - 'standard' (default)
                Computes the surface areas and perimeters by simply
                counting voxels. This is *much* faster but does not
                properly account for the rough voxelated nature
                of the surfaces.

            - 'high'
                Computes surface areas using the marching cube
                method, and perimeters using the fast marching method. These
                are substantially slower but better account for the
                voxelated nature of the images.

    voxel_size : scalar (default = 1)
        The resolution of the image, expressed as the length of one side
        of a voxel, so the volume of a voxel would be **voxel_size**-cubed.
    r_max : int
        The radius of the spherical structuring element to use in the
        Maximum filter stage that is used to find peaks. The default is 4.
    sigma : float
        The standard deviation of the Gaussian filter used in step 1. The
        default is 0.4.  If 0 is given then the filter is not applied.
    peaks : ndarray, optional
        Optionally, it is possible to supply an array containing peaks, which
        are used as markers in the watershed segmentation. If a boolean array
        is received (``True`` indicating peaks), then ``scipy.ndimage.label``
        with cubic connectivity is used to label them. If an integer array is
        received then it is assumed the peaks have already been labelled.
        This allows for comparison of peak finding algorithms for instance.
        If this argument is provided, then ``r_max`` and ``sigma`` are ignored
        since these are specfically used in the peak finding process. This
        array should contain peaks for all phases, and they are masked by
        the ``phases`` argument. If ``peaks`` are provided the parallelization
        is disabled.
    parallelization : dict
        The arguments for controlling the parallelization of the watershed
        function are rolled into this dictionary, otherwise the function
        signature would become too complex. Refer to the docstring of
        ``snow_partitioning_parallel`` for complete details. If no values
        are provided then the defaults for that function are used here.
        To disable parallelization pass ``parallel=None``, which will
        invoke the standard ``snow_partitioning`` or ``snow_partitioning_n``.
        If ``peaks`` are provided the parallelization is disabled.

    Returns
    -------
    network : Results object
        A custom object is returned with the following data added as attributes:

        - 'phases'
            The original ``phases`` image with any padding applied

        - 'regions'
            The watershed segmentation of the image, including boundary
            regions if padding was applied

        - 'network'
            A dictionary containing all the extracted network properties in
            OpenPNM format ('pore.coords', 'throat.conns', etc).

    References
    ----------
    .. [1] Gostick JT. Versatile and efficient pore network extraction
       method using marker-based watershed segmentation. Phys. Rev. E. 96,
       023307 (2017)
    .. [2] Khan ZA, Tranter TG, Agnaou M, Elkamel A, and Gostick JT, Dual
       network extraction algorithm to investigate multiple transport
       processes in porous materials: Image-based modeling of pore and
       grain-scale processes. Computers and Chemical Engineering. 123(6),
       64-77 (2019)
    .. [3] Khan ZA, García-Salaberri PA, Heenan T, Jervis R, Shearing P,
       Brett D, Elkamel A, Gostick JT, Probing the structure-performance
       relationship of lithium-ion battery cathodes using pore-networks
       extracted from three-phase tomograms. Journal of the
       Electrochemical Society. 167(4), 040528 (2020)
    .. [4] Khan ZA, Elkamel A, Gostick JT, Efficient extraction of pore
       networks from massive tomograms via geometric domain decomposition.
       Advances in Water Resources. 145(Nov), 103734 (2020)

    Examples
    --------
    `Click here
    <https://porespy.org/examples/networks/reference/snow2.html>`_
    to view online example.

    """
    # Parallel snow does not accept peaks, so if they are provided,
    # disable parallelization
    phases = phases.astype(int)
    if phase_alias is not None:
        vals = phase_alias.keys()
    else:
        vals = np.unique(phases)
        vals = vals[vals > 0]
    if peaks is not None:
        parallelization = None
    regions = None
    for i in vals:
        logger.info(f"Processing phase {i}...")
        phase = phases == i
        pk = None if peaks is None else peaks*phase
        if parallelization is not None:
            snow = snow_partitioning_parallel(
                im=phase, sigma=sigma, r_max=r_max, **parallelization)
        else:
            snow = snow_partitioning(im=phase, sigma=sigma, r_max=r_max,
                                     peaks=pk)
        if regions is None:
            regions = np.zeros_like(snow.regions, dtype=int)
        # Note: Using snow.regions > 0 here instead of phase is needed to
        # handle a bug in snow_partitioning, see issue #169 and #430
        regions += snow.regions + regions.max()*(snow.regions > 0)
    if phases.shape != regions.shape:
        logger.warning(f"Image was cropped to {regions.shape} during watershed")
        for ax in range(phases.ndim):
            phases = np.swapaxes(phases, 0, ax)
            phases = phases[:regions.shape[ax], ...]
            phases = np.swapaxes(phases, 0, ax)
    # Inspect and clean-up boundary_width argument
    boundary_width = _parse_pad_width(boundary_width, phases.shape)
    # If boundaries were specified, pad the images accordingly
    if np.any(boundary_width):
        regions = add_boundary_regions(regions, pad_width=boundary_width)
        phases = np.pad(phases, pad_width=boundary_width, mode='edge')
    # Perform actual extractcion on all regions
    net = regions_to_network(
        regions, phases=phases, accuracy=accuracy, voxel_size=voxel_size)
    # If image is multiphase, label pores/throats accordingly
    if phases.max() > 1:
        phase_alias = _parse_phase_alias(phase_alias, phases)
        net = label_phases(net, alias=phase_alias)
    # If boundaries were added, label them accordingly
    if np.any(boundary_width):
        W = boundary_width.flatten()
        L = ['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax'][:phases.ndim*2]
        L = [L[i]*int(W[i] > 0) for i in range(len(L))]
        L = np.reshape(L, newshape=boundary_width.shape)
        net = label_boundaries(net, labels=L)
    result = Results()
    result.network = net
    result.regions = regions
    result.phases = phases
    return result


def _parse_phase_alias(alias, phases):
    r"""
    """
    if alias is None:
        alias = {i+1: 'phase' + str(i+1) for i in range(phases.max())}
    for i in range(phases.max()):
        if i+1 not in alias.keys():
            alias[i+1] = 'phase'+str(i+1)
    return alias


def _parse_pad_width(pad_width, shape):
    r"""
    """
    ndim = len(shape)
    pad_width = np.atleast_1d(np.array(pad_width, dtype=object))

    if np.size(pad_width) == 1:
        pad_width = np.tile(pad_width.item(), ndim).astype(object)
    if len(pad_width) != ndim:
        raise Exception(f"pad_width must be scalar or {ndim}-element list")

    tmp = []
    for elem in pad_width:
        if np.size(elem) == 1:
            tmp.append(np.tile(np.array(elem).item(), 2))
        elif np.size(elem) == 2 and np.ndim(elem) == 1:
            tmp.append(elem)
        else:
            raise Exception("pad_width components can't have 2+ elements")

    return np.array(tmp, dtype=int)



import logging
import numpy as np
from edt import edt
from porespy.tools import get_tqdm, get_border
from porespy.tools import Results


__all__ = [
    "ibip_gpu",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


def ibip_gpu(im, dt=None, inlets=None, maxiter=10000):  # pragma: no cover
    """
    Performs invasion percolation on given image using iterative image
    dilation on GPU.

    Parameters
    ----------
    im : array_like
        Boolean array with ``True`` values indicating void voxels. If a
        standard numpy array is passed, it is converted to a cupy array.
    dt : array_like, optional
        The distance transform of ``im``. If a standard numpy array is
        passed, it is converted to a cupy array.
    inlets : array_like, optional
        Boolean array with ``True`` values indicating where the invading
        fluid is injected from.  If ``None``, all faces will be used.
        If a standard numpy array is passed, it is converted to a cupy
        array.
    maxiter : int, optional
        The number of steps to apply before stopping.  The default is to
        run for 10,000 steps which is almost certain to reach completion
        if the image is smaller than about 250-cubed.

    Returns
    -------
    results : Results object
        A custom object with the following two arrays as attributes:

        'inv_sequence'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the sequence at which it was invaded.

        'inv_size'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the ``inv_size`` at which was filled.

    """
    import cupy as cp
    from cupyx.scipy import ndimage as cndi

    im_gpu = cp.array(im)
    dt = edt(cp.asnumpy(im)) if dt is None else dt
    dt_gpu = cp.array(dt)
    inlets = get_border(shape=im.shape) if inlets is None else inlets
    inlets_gpu = cp.array(inlets)
    bd_gpu = cp.copy(inlets_gpu > 0)
    dt_gpu = dt_gpu.astype(int)

    # Alternative to _ibip
    inv_gpu = -1*((~im_gpu).astype(int))
    sizes_gpu = -1*((~im_gpu).astype(int))
    strel_gpu = ball_gpu if im_gpu.ndim == 3 else disk_gpu

    for step in tqdm(range(1, maxiter)):
        temp_gpu = cndi.binary_dilation(input=bd_gpu,
                                        structure=strel_gpu(1, smooth=False))
        edge_gpu = temp_gpu * (dt_gpu > 0)
        if ~cp.any(edge_gpu):
            logger.info('No more accessible invasion sites found')
            break
        # Find the maximum value of the dt underlaying the new edge
        r_max_gpu = dt_gpu[edge_gpu].max()
        # Find all values of the dt with that size
        dt_thresh_gpu = dt_gpu >= r_max_gpu
        # Insert the disk/sphere
        pt_gpu = cp.where(edge_gpu * dt_thresh_gpu)  # will be used later in updating bd
        # Update inv image
        bi_dial_gpu = cndi.binary_dilation(input=edge_gpu*dt_thresh_gpu,
                                           structure=strel_gpu(r_max_gpu.item()))
        bi_dial_step_gpu = bi_dial_gpu * step
        inv_prev_gpu = cp.copy(inv_gpu)
        mask_inv_prev_gpu = ~(inv_prev_gpu > 0)
        dial_single_gpu = mask_inv_prev_gpu * bi_dial_step_gpu
        inv_gpu = inv_prev_gpu + dial_single_gpu
        # Update size image
        bi_dial_size_gpu = bi_dial_gpu * r_max_gpu
        sizes_prev_gpu = cp.copy(sizes_gpu)
        mask_sizes_prev_gpu = ~(sizes_prev_gpu > 0)
        dial_single_size_gpu = mask_sizes_prev_gpu * bi_dial_size_gpu
        sizes_gpu = sizes_prev_gpu + dial_single_size_gpu
        # Update boundary image with newly invaded points
        bd_gpu[pt_gpu] = True
        dt_gpu[pt_gpu] = 0
        if step == (maxiter - 1):  # If max_iters reached, end loop
            logger.info('Maximum number of iterations reached')
            break

    temp_gpu = inv_gpu == 0
    inv_gpu[~im_gpu] = 0
    inv_gpu[temp_gpu] = -1
    inv_seq_gpu = make_contiguous_gpu(im=inv_gpu)
    temp_gpu = sizes_gpu == 0
    sizes_gpu[~im_gpu] = 0
    sizes_gpu[temp_gpu] = -1
    inv_sequence = cp.asnumpy(inv_seq_gpu)
    inv_size = cp.asnumpy(sizes_gpu)
    results = Results()
    results.inv_sequence = inv_sequence
    results.inv_size = inv_size
    return results


def rankdata_gpu(im_arr):  # pragma: no cover
    """
    GPU alternative to scipy's rankdata using 'dense' method.
    Assign ranks to data, dealing with ties appropriately.

    Parameters
    ----------
    im_arr : cupy ndarray
        Input image.

    Returns
    -------
    dense : cupy ndarray
        An array of length equal to the size of im_arr, containing rank
        scores.

    """
    import cupy as cp
    arr = cp.ravel(im_arr)
    sorter = cp.argsort(arr)
    inv = cp.empty(sorter.size, dtype=cp.intp)
    inv[sorter] = cp.arange(sorter.size, dtype=cp.intp)
    arr = arr[sorter]
    obs = cp.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv]
    return dense


def make_contiguous_gpu(im):  # pragma: no cover
    """
    Take an image with arbitrary greyscale values and adjust them to
    ensure all values fall in a contiguous range starting at 0.

    Parameters
    ----------
    im : cupy ndarray
        Input array containing greyscale values

    Returns
    -------
    im_new : cupy ndarray
        Array the same size as ``im`` but with all values in contiguous
        order.

    """
    shape = im.shape
    im_flat = im.flatten()
    mask_neg = im_flat < 0
    im_neg = -rankdata_gpu(-im_flat[mask_neg])
    mask_pos = im_flat > 0
    im_pos = rankdata_gpu(im_flat[mask_pos])
    im_flat[mask_pos] = im_pos
    im_flat[mask_neg] = im_neg
    im_new = np.reshape(im_flat, shape)
    return im_new


def ball_gpu(radius, smooth=True):  # pragma: no cover
    """
    Generates a ball-shaped structuring element.

    Parameters
    ----------
    radius : int
        The radius of the ball-shaped structuring element.
    smooth : bool, optional
        Indicates whether the balls should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    cupy ndarray
        The structuring element where elements of the neighborhood
        are 1 and 0 otherwise.

    """
    import cupy as cp
    n = 2 * radius + 1
    Z, Y, X = cp.mgrid[-radius:radius:n * 1j,
                       -radius:radius:n * 1j,
                       -radius:radius:n * 1j]
    s = X ** 2 + Y ** 2 + Z ** 2
    if smooth:
        radius = radius - 0.001
    return s <= radius * radius


def disk_gpu(radius, smooth=True):  # pragma: no cover
    """
    Generates a flat, disk-shaped structuring element.

    Parameters
    ----------
    radius : int
        The radius of the disk-shaped structuring element.
    smooth : bool, optional
        Indicates whether the disks should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    cupy ndarray
        The structuring element where elements of the neighborhood are
        1 and 0 otherwise.

    """
    import cupy as cp
    L = cp.arange(-radius, radius + 1)
    X, Y = cp.meshgrid(L, L)
    if smooth:
        radius = radius - 0.001
    return (X ** 2 + Y ** 2) <= radius ** 2


if __name__ == '__main__':
    import porespy as ps
    im = ps.generators.blobs(shape=[200, 200])
    out = ps.filters.ibip_gpu(im=im)



import logging
import sys
import numpy as np
import importlib
from dataclasses import dataclass
import psutil
import inspect
import time


logger = logging.getLogger("porespy")


__all__ = [
    'sanitize_filename',
    'get_tqdm',
    'show_docstring',
    'Results',
]


def _is_ipython_notebook():  # pragma: no cover
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True     # Jupyter notebook or qtconsole
        if shell == 'TerminalInteractiveShell':
            return False    # Terminal running IPython
        return False        # Other type (?)
    except NameError:
        return False        # Probably standard Python interpreter


@dataclass
class Settings:  # pragma: no cover
    r"""
    A dataclass for use at the module level to store settings.  This class
    is defined as a Singleton so now matter how or where it gets
    instantiated the same object is returned, containing all existing
    settings.

    Parameters
    ----------
    notebook : boolean
        Is automatically determined upon initialization of PoreSpy, and is
        ``True`` if running within a Jupyter notebook and ``False``
        otherwise. This is used by the ``porespy.tools.get_tqdm`` function
        to determine whether a standard or a notebook version of the
        progress bar should be used.
    tqdm : dict
        This dictionary is passed directly to the the ``tqdm`` function
        throughout PoreSpy (``for i in tqdm(range(N), **settings.tqdm)``).
        To see a list of available options visit the tqdm website.
        Probably the most important is ``'disable'`` which when set to
        ``True`` will silence the progress bars.  It's also possible to
        adjust the formatting such as ``'colour'`` and ``'ncols'``, which
        controls width.
    loglevel : str, or int
        Determines what messages to get printed in console. Options are:
        "TRACE" (5), "DEBUG" (10), "INFO" (20), "SUCCESS" (25), "WARNING" (30),
        "ERROR" (40), "CRITICAL" (50)

    """
    __instance__ = None
    # Might need to add 'file': sys.stdout to tqdm dict
    tqdm = {'disable': False,
            'colour': None,
            'ncols': None,
            'leave': False,
            'file': sys.stdout}
    _loglevel = 40 if _is_ipython_notebook() else 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._notebook = None
        self._ncores = psutil.cpu_count(logical=False)

    @property
    def loglevel(self):
        return self._loglevel

    @loglevel.setter
    def loglevel(self, value):
        if isinstance(value, str):
            options = {
                "TRACE" : 5,
                "DEBUG" : 10,
                "INFO" : 20,
                "SUCESS" : 25,
                "WARNING" : 30,
                "ERROR" : 40,
                "CRITICAL" : 50
            }
            value = options[value]
        self._loglevel = value
        logger.setLevel(value)

    def __new__(cls):
        if Settings.__instance__ is None:
            Settings.__instance__ = super().__new__(cls)
        return Settings.__instance__

    def __repr__(self):
        indent = 0
        for item in self.__dir__():
            if not item.startswith('_'):
                indent = max(indent, len(item) + 1)
        s = ''
        for item in self.__dir__():
            if not item.startswith('_'):
                s += ''.join((item, ':', ' '*(indent-len(item))))
                attr = getattr(self, item)
                temp = ''.join((attr.__repr__(), '\n'))
                if isinstance(attr, dict):
                    temp = temp.replace(',', '\n' + ' '*(indent + 1))
                s += temp
        return s

    def _get_ncores(self):
        if self._ncores is None:
            self._ncores = psutil.cpu_count(logical=False)
        return self._ncores

    def _set_ncores(self, val):
        cpu_count = psutil.cpu_count(logical=False)
        if val is None:
            val = cpu_count
        elif val > cpu_count:
            logger.error('Value is more than the available number of cores')
            val = cpu_count
        self._ncores = val

    ncores = property(fget=_get_ncores, fset=_set_ncores)

    def _get_notebook(self):
        if self._notebook is None:
            self._notebook = _is_ipython_notebook()
        return self._notebook

    def _set_notebook(self, val):
        logger.error('This value is determined automatically at runtime')

    notebook = property(fget=_get_notebook, fset=_set_notebook)


def get_tqdm():  # pragma: no cover
    r"""
    Fetches a version of ``tqdm`` function that depends on the environment.

    Either text-based for the IPython console or gui-based for Jupyter
    notebooks.

    Returns
    -------
    tqdm : function handle
        The function to use when wrapping an iterator (i.e. tqdm(range(n)))

    """
    if Settings().notebook is True:
        tqdm = importlib.import_module('tqdm.notebook')
    else:
        tqdm = importlib.import_module('tqdm')
    return tqdm.tqdm


def show_docstring(func):  # pragma: no cover
    r"""
    Fetches the docstring for a function and returns it in markdown format.

    Useful for printing in a Jupyter notebook.

    Parameters
    ----------
    func : object
        Function handle to function whose docstring is desired

    Returns
    -------
    md : str
        A string with the markdown syntax included, suitable for printing
        in a Jupyter notebook using the ``IPython.display.Markdown``
        function.

    """
    title = f'---\n ## Documentation for ``{func.__name__}``\n ---\n'
    try:
        from npdoc_to_md import render_md_from_obj_docstring
        txt = render_md_from_obj_docstring(obj=func, obj_namespace=func.__name__)
    except ModuleNotFoundError:
        txt = func.__doc__
    return title + txt + '\n---'


def sanitize_filename(filename, ext, exclude_ext=False):
    r"""
    Returns a sanitized string in the form of name.extension

    Parameters
    ----------
    filename : str
        Unsanitized filename, could be 'test.vtk' or just 'test'

    ext : str
        Extension of the file, could be 'vtk'

    exclude_ext : bool
        If True, the returned string doesn't have the extension

    Returns
    -------
    sanitized : str
        Sanitized filename in form of name.extension

    """
    ext.strip(".")
    if filename.endswith(f".{ext}"):
        name = ".".join(filename.split(".")[:-1])
    else:
        name = filename
    filename_formatted = f"{name}" if exclude_ext else f"{name}.{ext}"
    return filename_formatted


class Results:
    r"""
    A minimal class for use when returning multiple values from a function

    This class supports dict-like assignment and retrieval
    (``obj['im'] = im``), namedtuple-like attribute look-ups (``obj.im``),
    and generic class-like object assignment (``obj.im = im``)

    """

    def __init__(self, **kwargs):
        self._func = inspect.getouterframes(inspect.currentframe())[1].function
        self._time = time.asctime()

    def __iter__(self):
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                yield v

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __str__(self):
        header = "―" * 78
        lines = [
            header,
            f"Results of {self._func} generated at {self._time}",
            header,
        ]
        for item in list(self.__dict__.keys()):
            if item.startswith('_'):
                continue
            if (isinstance(self[item], np.ndarray)):
                s = np.shape(self[item])
                lines.append("{0:<25s} Array of size {1}".format(item, s))
            elif hasattr(self[item], 'keys'):
                N = len(self[item].keys())
                lines.append("{0:<25s} Dictionary with {1} items".format(item, N))
            else:
                lines.append("{0:<25s} {1}".format(item, self[item]))
        lines.append(header)
        return "\n".join(lines)



import os
import sys
import codecs
import os.path
from distutils.util import convert_path
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

sys.path.append(os.getcwd())
ver_path = convert_path('porespy/__version__.py')


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            ver = line.split(delim)[1].split(".")
            if "dev0" in ver:
                ver.remove("dev0")
            return ".".join(ver)
    else:
        raise RuntimeError("Unable to find version string.")


# Read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='porespy',
    description='A set of tools for analyzing 3D images of porous materials',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=get_version(ver_path),
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    packages=[
        'porespy',
        'porespy.tools',
        'porespy.generators',
        'porespy.metrics',
        'porespy.filters',
        'porespy.filters.imagej',
        'porespy.networks',
        'porespy.dns',
        'porespy.simulations',
        'porespy.visualization',
        'porespy.io'
    ],
    install_requires=[
        'dask',
        'deprecated',
        'edt',
        'matplotlib',
        'numba',
        'numpy',
        'openpnm',
        'pandas',
        'psutil',
        'rich',
        'scikit-image',
        'scipy',
        'tqdm',
    ],
    author='PoreSpy Team',
    author_email='jgostick@gmail.com',
    download_url='https://github.com/PMEAL/porespy/',
    url='http://porespy.org',
    project_urls={
        'Documentation': 'https://porespy.org/',
        'Source': 'https://github.com/PMEAL/porespy/',
        'Tracker': 'https://github.com/PMEAL/porespy/issues',
    },
)



import logging
import numpy as np
import scipy.ndimage as spim
from edt import edt
from skimage.morphology import disk, ball
from porespy import settings
from porespy.tools import get_tqdm, ps_round, get_border
from porespy.tools import _insert_disks_at_points
from porespy.filters import trim_disconnected_blobs, fftmorphology
# import random
from numba import njit


__all__ = [
    "pseudo_gravity_packing",
    "pseudo_electrostatic_packing",
]


tqdm = get_tqdm()
logger = logging.getLogger(__name__)


@njit
def _set_seed(a):
    np.random.seed(a)


def pseudo_gravity_packing(
    im,
    r,
    clearance=0,
    axis=0,
    maxiter=1000,
    edges='contained',
    seed=None,
):
    r"""
    Iteratively inserts spheres at the lowest accessible point in an image,
    mimicking a gravity packing.

    Parameters
    ----------
    im : ndarray
        Image with ``True`` values indicating the phase where spheres should be
        inserted. A common option would be a cylindrical plug which would
        result in a tube filled with beads.
    r : int
        The radius of the spheres to be added
    clearance : int (default is 0)
        The amount space to add between neighboring spheres. The value can be
        negative for overlapping spheres, but ``abs(clearance) > r``.
    axis : int (default is 0)
        The axis along which gravity acts.
    maxiter : int (default is 1000)
        The maximum number of spheres to add
    edges : string (default is 'contained')
        Controls how spheres at the edges of the image are handled.  Options are:

        ============ ================================================================
        edges        description
        ============ ================================================================
        'contained'  Spheres are all completely within the image
        'extended'   Spheres are allowed to extend beyond the edge of the
                     image.  In this mode the volume fraction will be less that
                     requested since some spheres extend beyond the image, but their
                     entire volume is counted as added for computational efficiency.
        ============ ================================================================

    seed : int, optional, default = `None`
        The seed to supply to the random number generator. Because this function
        uses ``numba`` for speed, calling the normal ``numpy.random.seed(<seed>)``
        has no effect. To get a repeatable image, the seed must be passed to the
        function so it can be initialized the way ``numba`` requires. The default
        is ``None``, which means each call will produce a new realization.

    Returns
    -------
    spheres : ndarray
        An image the same size as ``im`` with spheres indicated by ``True``.
        The spheres are only inserted at locations that are accessible
        from the top of the image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/pseudo_gravity_packing.html>`_
    to view online example.

    """
    logger.debug(f'Adding spheres of radius {r}')

    if seed is not None:  # Initialize rng so numba sees it
        _set_seed(seed)
        np.random.seed(seed)

    im = np.swapaxes(im, 0, axis)
    im_temp = np.zeros_like(im, dtype=bool)
    r = r - 1
    strel = disk if im.ndim == 2 else ball
    sites = fftmorphology(im == 1, strel=strel(r), mode='erosion')
    inlets = np.zeros_like(im)
    inlets[-(r+1), ...] = True
    sites = trim_disconnected_blobs(im=sites, inlets=inlets)
    x_min = np.where(sites)[0].min()
    n = None
    for n in tqdm(range(maxiter), **settings.tqdm):
        if im.ndim == 2:
            x, y = np.where(sites[x_min:x_min+2*r, ...])
        else:
            x, y, z = np.where(sites[x_min:x_min+2*r, ...])
        if len(x) == 0:
            break
        options = np.where(x == x.min())[0]
        choice = np.random.randint(len(options))
        if im.ndim == 2:
            cen = np.vstack([x[options[choice]] + x_min,
                             y[options[choice]]])
        else:
            cen = np.vstack([x[options[choice]] + x_min,
                             y[options[choice]],
                             z[options[choice]]])
        im_temp = _insert_disks_at_points(im_temp, coords=cen,
                                          radii=np.array([r - clearance]),
                                          v=True, overwrite=True)
        sites = _insert_disks_at_points(sites, coords=cen,
                                        radii=np.array([2*r]),
                                        v=0,
                                        overwrite=True)
        x_min += x.min()
    logger.debug(f'A total of {n} spheres were added')
    im_temp = np.swapaxes(im_temp, 0, axis)
    return im_temp


def pseudo_electrostatic_packing(
    im,
    r,
    sites=None,
    clearance=0,
    protrusion=0,
    edges='extended',
    maxiter=1000,
    seed=None,
):
    r"""
    Iterativley inserts spheres as close to the given sites as possible.

    Parameters
    ----------
    im : ndarray
        Image with ``True`` values indicating the phase where spheres should be
        inserted.
    r : int
        Radius of spheres to insert.
    sites : ndarray (optional)
        An image with ``True`` values indicating the electrostatic attraction
        points.
        If this is not given then the peaks in the distance transform are used.
    clearance : int (optional, default=0)
        The amount of space to put between each sphere. Negative values are
        acceptable to create overlaps, but ``abs(clearance) < r``.
    protrusion : int (optional, default=0)
        The amount that spheres are allowed to protrude beyond the active phase.
    maxiter : int (optional, default=1000)
        The maximum number of spheres to insert.
    edges : string (default is 'contained')
        Controls how spheres at the edges of the image are handled.  Options are:

        ============ ================================================================
        edges        description
        ============ ================================================================
        'contained'  Spheres are all completely within the image
        'extended'   Spheres are allowed to extend beyond the edge of the
                     image.  In this mode the volume fraction will be less that
                     requested since some spheres extend beyond the image, but their
                     entire volume is counted as added for computational efficiency.
        ============ ================================================================

    seed : int, optional, default = `None`
        The seed to supply to the random number generator. Because this function
        uses ``numba`` for speed, calling the normal ``numpy.random.seed(<seed>)``
        has no effect. To get a repeatable image, the seed must be passed to the
        function so it can be initialized the way ``numba`` requires. The default
        is ``None``, which means each call will produce a new realization.

    Returns
    -------
    im : ndarray
        An image with inserted spheres indicated by ``True``

    Examples
    --------
    `Click here
    <https://porespy.org/examples/generators/reference/pseudo_electrostatic_packing.html>`_
    to view online example.

    """
    if seed is not None:  # Initialize rng so numba sees it
        _set_seed(seed)
        np.random.seed(seed)

    im_temp = np.zeros_like(im, dtype=bool)
    dt_im = edt(im)
    if sites is None:
        dt2 = spim.gaussian_filter(dt_im, sigma=0.5)
        strel = ps_round(r, ndim=im.ndim, smooth=True)
        sites = (spim.maximum_filter(dt2, footprint=strel) == dt2)*im
    dt = edt(sites == 0).astype(int)
    sites = (sites == 0)*(dt_im >= (r - protrusion))
    if dt_im.max() < np.inf:
        dtmax = int(dt_im.max()*2)
    else:
        dtmax = min(im.shape)
    dt[~sites] = dtmax
    if edges == 'contained':
        borders = get_border(im.shape, thickness=r, mode='faces')
        dt[borders] = dtmax
    r = r + clearance
    # Get initial options
    options = np.where(dt == 1)
    for _ in tqdm(range(maxiter), **settings.tqdm):
        hits = dt[options] < dtmax
        if hits.sum(dtype=np.int64) == 0:
            if dt.min() == dtmax:
                break
            options = np.where(dt == dt.min())
            hits = dt[options] < dtmax
        if hits.size == 0:
            break
        choice = np.random.choice(np.where(hits)[0])
        cen = np.vstack([options[i][choice] for i in range(im.ndim)])
        im_temp = _insert_disks_at_points(im_temp, coords=cen,
                                          radii=np.array([r-clearance]),
                                          v=True,
                                          overwrite=True)
        dt = _insert_disks_at_points(dt, coords=cen,
                                     radii=np.array([2*r-clearance]),
                                     v=int(dtmax),
                                     overwrite=True)
    return im_temp


if __name__ == "__main__":
    import porespy as ps
    import matplotlib.pyplot as plt
    shape = [200, 200]

    fig, ax = plt.subplots(1, 2)
    im = ps.generators.pseudo_gravity_packing(im=np.ones(shape, dtype=bool),
                                              r=7, clearance=3,
                                              edges='contained', seed=0)
    ax[0].imshow(im, origin='lower')

    sites = np.zeros(shape, dtype=bool)
    sites[100, 100] = True
    im = ps.generators.pseudo_electrostatic_packing(im=np.ones(shape,
                                                               dtype=bool),
                                                    r=5, sites=sites,
                                                    clearance=4,
                                                    maxiter=50, seed=0)
    ax[1].imshow(im, origin='lower')



import porespy as ps
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from matplotlib import animation
from copy import copy
from porespy import settings


__all__ = [
    'set_mpl_style',
    'satn_to_movie',
    'satn_to_panels',
    'prep_for_imshow',
]


def set_mpl_style():  # pragma: no cover
    r"""
    Prettifies matplotlib's output by adjusting fonts, markersize etc.
    """
    sfont = 12
    mfont = 12
    lfont = 12

    image_props = {'interpolation': 'none',
                   'cmap': 'viridis'}
    line_props = {'linewidth': 2,
                  'markersize': 8,
                  'markerfacecolor': 'w'}
    font_props = {'size': sfont}
    axes_props = {'titlesize': lfont,
                  'labelsize': mfont,
                  'linewidth': 2,
                  'labelpad': 8}
    xtick_props = {'labelsize': sfont,
                   'top': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    ytick_props = {'labelsize': sfont,
                   'right': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    legend_props = {'fontsize': mfont,
                    'frameon': False}
    figure_props = {'titlesize': sfont,
                    'autolayout': True}

    plt.rc('font', **font_props)
    plt.rc('lines', **line_props)
    plt.rc('axes', **axes_props)
    plt.rc('xtick', **xtick_props)
    plt.rc('ytick', **ytick_props)
    plt.rc('legend', **legend_props)
    plt.rc('figure', **figure_props)
    plt.rc('image', **image_props)

    if ps.settings.notebook:
        import IPython
        IPython.display.set_matplotlib_formats('retina')


def satn_to_movie(im, satn, cmap='viridis',
                  c_under='grey', c_over='white',
                  v_under=1e-3, v_over=1.0, fps=10, repeat=True):
    r"""
    Converts a saturation map into an animation that can be saved

    Parameters
    ----------
    im : ndarray
        The boolean image of the porous media with ``True`` values indicating
        the void space
    satn : ndaray
        The saturation map such as that produced by an invasion or drainage
        algorithm.
    cmap : str
        The name of the matplotlib color map to use. These are listed on
        matplotlib's website
        `here <https://matplotlib.org/stable/gallery/color/colormap_reference.html>`__
    c_under, c_over : str
        The color to insert for values that are less than `v_under`
        (greater than `v_over`).  The string value of colors are given on
        matplotlib's website
        `here <https://matplotlib.org/stable/gallery/color/named_colors.html>`__
    v_under, v_over : scalar
        The values in ``satn`` that should be considered the lower and upper
        threshold, beyond which the colors given in `c_under` and `c_over`
        are used.
    fps : int
        The frames per second to use when generating the movie.  A higher
        number gives a shorter and faster-paced movie.
    repeat : bool
        If ``True`` the produced animation will rerun repeatedly until
        stopped or closed.

    Notes
    -----
    To save animation as a file use:
    ``ani.save('image_based_ip.gif', writer='imagemagick', fps=3)``

    Examples
    --------
    `Click here
    <https://porespy.org/examples/visualization/reference/satn_to_movie.html>`_
    to view online example.
    """
    # Define nice color map
    cmap = copy(plt.cm.get_cmap(name=cmap))
    cmap.set_over(color=c_over)
    cmap.set_under(color=c_under)

    # Reduce inv_satn image to limited number of values to speed-up movie
    target = np.around(satn, decimals=3)
    seq = np.zeros_like(target)  # Empty image to place frame
    movie = []  # List to append each frame
    fig, ax = plt.subplots(1, 1)
    steps = np.unique(target)[1:]
    with tqdm(steps, **settings.tqdm) as pbar:
        for v in steps:
            pbar.update()
            seq += v*(target == v)
            seq[~im] = target.max() + 10
            frame1 = ax.imshow(seq, vmin=v_under, vmax=v_over,
                               animated=True, cmap=cmap, origin='lower',
                               interpolation='none')
            movie.append([frame1])
    ani = animation.ArtistAnimation(fig, movie, interval=int(1000/fps),
                                    blit=True, repeat=repeat,
                                    repeat_delay=1.0)
    return ani


def satn_to_panels(satn, im, bins=None, axis=0, slice=None, **kwargs):
    r"""
    Produces a set of images with each panel containing one saturation

    Parameters
    ----------
    satn : ndarray
        An image with each voxel indicating the global saturation at which
        it was invaded.  0 indicates solid and -1 indicates uninvaded.
    im : ndarray
        A boolean image with ``True`` values indicating the void voxels and
        ``False`` for solid.
    bins : int
        Indicates for which saturations images should be made. If an ``int``
        then a list of equally space values between 0 and 1 is generated.
        If ``None`` (default) than all saturation values in the image are used.
    axis : int, optional
        If the image is 3D, a 2D image is extracted at the specified
        ``slice`` taken along this axis. If the image is 2D this is ignored.
    slice : int, optional
        If the image is 3D, a 2D image is extracted from this slice
        along the given ``axis``.  If ``None``, then a slice at the mid-point
        of the axis is returned.  If 2D this is ignored.
    **kwargs : various
        Additional keyword arguments are sent to the ``imshow`` function,
        such as ``interpolation``.

    Returns
    -------
    fig, ax : Matplotlib figure and axis objects
        The same things as returned by ``plt.subplots``

    Examples
    --------
    `Click here
    <https://porespy.org/examples/visualization/reference/satn_to_panels.html>`_
    to view online example.
    """
    def factors(n):
        return sorted(list(set(
            factor for i in range(1, int(n**0.5) + 1) if n % i == 0
            for factor in (i, n//i)
        )))

    if bins is None:
        Ps = np.unique(satn)
    elif isinstance(bins, int):
        Ps = np.linspace(0, 1, bins+1)[1:]
    Ps = Ps[Ps > 0]
    f = factors(len(Ps))
    if len(Ps) < 4:
        m = 1
        n = len(Ps)
    elif len(f) % 2 == 0:
        m, n = f[int(len(f)/2-1)], f[int(len(f)/2)]
    else:
        m = f[int(len(f)/2)]
        n = m
    fig, ax = plt.subplots(m, n)
    ax = np.atleast_2d(ax)
    temp_old = np.zeros_like(im)
    for i, p in enumerate(Ps):
        temp = (satn <= p)*(satn > 0)
        im_data = prep_for_imshow(im=temp*2.0 - temp_old*1.0, mask=im,
                                  axis=axis, slice=slice)
        im_data.pop('vmax')
        [im_data.pop(i) for i in kwargs]
        ax[i // n][i % n].imshow(**im_data, vmax=2, **kwargs)
        ax[i // n][i % n].set_title(
            str(np.around(temp.sum(dtype=np.int64)/im.sum(dtype=np.int64),
                          decimals=5)))
        temp_old = np.copy(temp)
    return fig, ax


def prep_for_imshow(im, mask=None, axis=0, slice=None):
    r"""
    Adjusts the range of greyscale values in an image to improve visualization
    by ``matplotlib.pyplot.imshow``

    Parameters
    ----------
    im : ndimage
        The image to show. If ``im`` includes ``+inf`` or ``-inf`` values,
        they are converted to 1 above or below the minimum and maximum finite
        values in ``im``, respectively.
    mask : ndimage, optional
        An image of the porous material with ``True`` indicating voxels of
        interest. The ``False`` voxels are excluded from the ``vmax`` and
        ``vmin`` calculation.
    axis : int, optional
        If the image is 3D, a 2D image is returned with the specified
        ``slice`` taken along this axis (default = 0).  If ``None`` then a 3D
        image is returned. If the image is 2D this is ignored.
    slice : int, optional
        If ``im`` is 3D, a 2D image is be returned showing this slice
        along the given ``axis``.  If ``None``, then a slice at the mid-point
        of the axis is returned.  If ``axis`` is ``None`` or the image is 2D
        this is ignored.

    Returns
    -------
    kwargs : dict
        A python dicionary designed to be passed directly to
        ``matplotlib.pyplot.imshow`` using the "\*\*kwargs" features (i.e.
        ``plt.imshow(\*\*data)``).  It contains the following key-value pairs:

        =============== =======================================================
        key               value
        =============== =======================================================
        'X'             The adjusted image with ``+inf`` replaced by
                        ``vmax + 1``, and all solid voxels replacd by
                        ``np.nan`` to show as white in ``imshow``
        'vmax'          The maximum of ``values`` not including ``+inf`` or
                        values in ``False`` voxels in ``mask``.
        'vmin'          The minimum of ``values`` not including ``-inf`` or
                        values in ``False`` voxels in ``mask``.
        'interpolation' Set to 'none' to avoid artifacts in ``imshow``
        'origin'        Set to 'lower' to put (0, 0) on the bottom-left corner
        =============== =======================================================

    Notes
    -----
    If any of the *extra* items are unwanted they can be removed with
    ``del data['interpolation']`` or ``data.pop('interpolation')``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/visualization/reference/prep_for_imshow.html>`_
    to view online example.

    """
    # If 3D, fetch 2D slice immediately to save memory
    if (im.ndim == 3) and (axis is not None):
        if slice is None:
            slice = int(im.shape[axis]/2)
        # Rotate image to put given axis first, then slice
        im = np.swapaxes(im, 0, axis)[slice, ...]
        if mask is not None:  # Rotate mask as well, then slice
            mask = np.swapaxes(mask, 0, axis)[slice, ...]
    im = im.astype(float)
    if mask is None:
        mask = np.ones_like(im, dtype=bool)

    vmax = np.amax((im*(im < np.inf))[mask])
    im[(im == np.inf)] = vmax + 1
    vmin = np.amin((im*(im > -np.inf))[mask])
    im[(im == -np.inf)] = vmin - 1

    return {'X': im, 'vmin': vmin, 'vmax': vmax,
            'interpolation': 'none', 'origin': 'lower'}



import imageio
import numpy as np
import os
from zipfile import ZipFile
from porespy.tools import get_tqdm
from pathlib import Path


tqdm = get_tqdm()


__all__ = [
    'zip_to_stack',
    'folder_to_stack',
]


def folder_to_stack(target_dir):
    r"""
    Opens all images found in the target directory as single 3D numpy array

    Parameters
    ----------
    target_dir : str or path object
        The location of the folder containing the images.

    Returns
    -------
    im : ndarray
        A 3D numpy array of the imported image.

    Notes
    -----
    The files should be named with leading numerical values indicating their
    layer number, like 001, 002, etc.
    """
    p = Path(target_dir)
    test_im = imageio.v2.imread(os.path.join(p, os.listdir(p)[0]))
    im = np.zeros(shape=[test_im.shape[0],
                         test_im.shape[1],
                         len(os.listdir(p))],
                  dtype=test_im.dtype)
    for i, f in enumerate(tqdm(os.listdir(p))):
        im[..., i] = imageio.v2.imread(os.path.join(p, f))

    return im


def zip_to_stack(f):
    r"""
    Reads a zip file containing 2D slices of a 3D image, and converts to a 3D stack

    Parameters
    ----------
    f : str
        The path and/or file name of the zip archive.  If ``<name>.zip`` is given,
        then it's assumed to be located in the current working directory. Otherwise
        a full path should be given, like ``C:\path\to\file.zip``. Either way, the
        archive is extracted into a folder in the given directory.

    Returns
    -------
    im : ndarray
        A 3D numpy array of the imported image.

    Notes
    -----
    The files should be named with leading numerical values indicating their
    layer number, like 001, 002, etc.
    """
    p = Path(f)
    dir_for_files = p.parts[-1].rpartition('.')[0]

    with ZipFile(p, 'r') as f:
        f.extractall(dir_for_files)

    # Method 1: uses skimage and numpy function so is easy to understand
    # filenames = []
    # for f in os.listdir(target_dir):
    #     filenames.append(os.path.join(target_dir, f))
    # files = imread_collection(filenames, conserve_memory=True)
    # im = np.stack(files, axis=2)

    # Method 2: Same speed as 1 but more complex, but allows tqdm progress bar
    test_im = imageio.v2.imread(os.path.join(dir_for_files,
                                             os.listdir(dir_for_files)[0]))
    im = np.zeros(shape=[test_im.shape[0],
                         test_im.shape[1],
                         len(os.listdir(dir_for_files))],
                  dtype=test_im.dtype)
    for i, f in enumerate(tqdm(os.listdir(dir_for_files))):
        im[..., i] = imageio.v2.imread(os.path.join(dir_for_files , f))

    return im



import logging
import numpy as np
import openpnm as op
from porespy.filters import trim_nonpercolating_paths
from porespy.tools import Results
from porespy.generators import faces


logger = logging.getLogger(__name__)
ws = op.Workspace()


__all__ = ['tortuosity_fd']


def tortuosity_fd(im, axis, solver=None):
    r"""
    Calculates the tortuosity of image in the specified direction.

    Parameters
    ----------
    im : ndarray
        The binary image to analyze with ``True`` indicating phase of interest
    axis : int
        The axis along which to apply boundary conditions

    Returns
    -------
    results : Results object
        The following values are computed and returned as attributes:

        =================== ===================================================
        Attribute           Description
        =================== ===================================================
        tortuosity          Calculated using the ``effective_porosity`` as
                            :math:`\tau = \frac{D_{AB}}{D_{eff}} \cdot
                            \varepsilon`.
        effective_porosity  Porosity of the image after applying
                            ``trim_nonpercolating_paths``.  This removes
                            disconnected voxels which cause singular matrices.

        original_porosity   Porosity of the as-received the image

        formation_factor    Found as :math:`D_{AB}/D_{eff}`.

        concentration       An image containing the concentration values from
                            the simulation.
        =================== ===================================================

    Examples
    --------
    `Click here
    <https://porespy.org/examples/simulations/reference/tortuosity_fd.html>`_
    to view online example.

    """
    if axis > (im.ndim - 1):
        raise Exception(f"'axis' must be <= {im.ndim}")
    openpnm_v3 = op.__version__.startswith('3')

    # Obtain original porosity
    eps0 = im.sum(dtype=np.int64) / im.size

    # Remove floating pores
    inlets = faces(im.shape, inlet=axis)
    outlets = faces(im.shape, outlet=axis)
    im = trim_nonpercolating_paths(im, inlets=inlets, outlets=outlets)
    # Check if porosity is changed after trimmimg floating pores
    eps = im.sum(dtype=np.int64) / im.size
    if not eps:
        raise Exception('No pores remain after trimming floating pores')
    if eps < eps0:  # pragma: no cover
        logger.warning('Found non-percolating regions, were filled to percolate')

    # Generate a Cubic network to be used as an orthogonal grid
    net = op.network.CubicTemplate(template=im, spacing=1.0)
    if openpnm_v3:
        phase = op.phase.Phase(network=net)
    else:
        phase = op.phases.GenericPhase(network=net)
    phase['throat.diffusive_conductance'] = 1.0
    # Run Fickian Diffusion on the image
    fd = op.algorithms.FickianDiffusion(network=net, phase=phase)
    # Choose axis of concentration gradient
    inlets = net.coords[:, axis] <= 1
    outlets = net.coords[:, axis] >= im.shape[axis] - 1
    # Boundary conditions on concentration
    cL, cR = 1.0, 0.0
    fd.set_value_BC(pores=inlets, values=cL)
    fd.set_value_BC(pores=outlets, values=cR)
    if openpnm_v3:
        if solver is None:
            solver = op.solvers.PyamgRugeStubenSolver(tol=1e-8)
        fd._update_A_and_b()
        fd.x, info = solver.solve(fd.A.tocsr(), fd.b)
        if info:
            raise Exception(f'Solver failed to converge, exit code: {info}')
    else:
        fd.settings.update({'solver_family': 'scipy', 'solver_type': 'cg'})
        fd.run()

    # Calculate molar flow rate, effective diffusivity and tortuosity
    r_in = fd.rate(pores=inlets)[0]
    r_out = fd.rate(pores=outlets)[0]
    if not np.allclose(-r_out, r_in, rtol=1e-4):  # pragma: no cover
        logger.error(f"Inlet/outlet rates don't match: {r_in:.4e} vs. {r_out:.4e}")
    dC = cL - cR
    L = im.shape[axis]
    A = np.prod(im.shape) / L
    # L-1 because BCs are put inside the domain, see issue #495
    Deff = r_in * (L-1)/A / dC
    tau = eps / Deff

    # Attach useful parameters to Results object
    result = Results()
    result.im = im
    result.tortuosity = tau
    result.formation_factor = 1 / Deff
    result.original_porosity = eps0
    result.effective_porosity = eps
    conc = np.zeros(im.size, dtype=float)
    conc[net['pore.template_indices']] = fd['pore.concentration']
    result.concentration = conc.reshape(im.shape)

    # Free memory
    ws.close_project(net.project)

    return result



import numpy as np
from numba import njit


__all__ = [
    '_make_disk',
    '_make_disks',
    '_make_ball',
    '_make_balls',
    '_insert_disk_at_points',
    '_insert_disks_at_points',
]


@njit(parallel=False)
def _make_disks(r, smooth=True):  # pragma: no cover
    r"""
    Returns a list of disks from size 0 to ``r``

    Parameters
    ----------
    r : int
        The size of the largest disk to generate
    smooth : bool
        Indicates whether the disks should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    disks : list of ND-arrays
        A list containing the disk images, with the disk of radius R at index
        R of the list, meaning it can be accessed as ``disks[R]``.

    """
    disks = []
    for val in range(0, r):
        disk = _make_disk(val, smooth)
        disks.append(disk)
    return disks


@njit(parallel=False)
def _make_balls(r, smooth=True):  # pragma: no cover
    r"""
    Returns a list of balls from size 0 to ``r``

    Parameters
    ----------
    r : int
        The size of the largest ball to generate
    smooth : bool
        Indicates whether the balls should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    balls : list of ND-arrays
        A list containing the ball images, with the ball of radius R at index
        R of the list, meaning it can be accessed as ``balls[R]``.

    """
    balls = []
    for val in range(0, r):
        ball = _make_ball(val, smooth)
        balls.append(ball)
    return balls


@njit(parallel=False)
def _insert_disk_at_points(im, coords, r, v,
                           smooth=True, overwrite=False):  # pragma: no cover
    r"""
    Insert spheres (or disks) into the given ND-image at given locations

    This function uses numba to accelerate the process, and does not
    overwrite any existing values (i.e. only writes to locations containing
    zeros).

    Parameters
    ----------
    im : ND-array
        The image into which the spheres/disks should be inserted. This is an
        'in-place' operation.
    coords : ND-array
        The center point of each sphere/disk in an array of shape
        ``ndim by npts``
    r : int
        The radius of all the spheres/disks to add. It is assumed that they
        are all the same radius.
    v : scalar
        The value to insert
    smooth : boolean
        If ``True`` (default) then the spheres/disks will not have the litte
        nibs on the surfaces.

    """
    npts = len(coords[0])
    if im.ndim == 2:
        xlim, ylim = im.shape
        s = _make_disk(r, smooth)
        for i in range(npts):
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            if s[a, b] == 1:
                                if overwrite or (im[x, y] == 0):
                                    im[x, y] = v
    elif im.ndim == 3:
        xlim, ylim, zlim = im.shape
        s = _make_ball(r, smooth)
        for i in range(npts):
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            for c, z in enumerate(range(pt[2]-r, pt[2]+r+1)):
                                if (z >= 0) and (z < zlim):
                                    if (s[a, b, c] == 1):
                                        if overwrite or (im[x, y, z] == 0):
                                            im[x, y, z] = v
    return im


@njit(parallel=False)
def _insert_disks_at_points(im, coords, radii, v, smooth=True,
                            overwrite=False):  # pragma: no cover
    r"""
    Insert spheres (or disks) of specified radii into an ND-image at given locations.

    This function uses numba to accelerate the process, and does not overwrite
    any existing values (i.e. only writes to locations containing zeros).

    Parameters
    ----------
    im : ND-array
        The image into which the spheres/disks should be inserted. This is an
        'in-place' operation.
    coords : ND-array
        The center point of each sphere/disk in an array of shape
        ``ndim by npts``
    radii : array_like
        The radii of the spheres/disks to add.
    v : scalar
        The value to insert
    smooth : boolean, optional
        If ``True`` (default) then the spheres/disks will not have the litte
        nibs on the surfaces.
    overwrite : boolean, optional
        If ``True`` then the inserted spheres overwrite existing values.  The
        default is ``False``.

    """
    npts = len(coords[0])
    if im.ndim == 2:
        xlim, ylim = im.shape
        for i in range(npts):
            r = radii[i]
            s = _make_disk(r, smooth)
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            if s[a, b] == 1:
                                if overwrite or (im[x, y] == 0):
                                    im[x, y] = v
    elif im.ndim == 3:
        xlim, ylim, zlim = im.shape
        for i in range(npts):
            r = radii[i]
            s = _make_ball(r, smooth)
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            for c, z in enumerate(range(pt[2]-r, pt[2]+r+1)):
                                if (z >= 0) and (z < zlim):
                                    if s[a, b, c] == 1:
                                        if overwrite or (im[x, y, z] == 0):
                                            im[x, y, z] = v
    return im


@njit(parallel=False)
def _make_disk(r, smooth=True):  # pragma: no cover
    r"""
    Generate a circular structuring element of the given radius

    Parameters
    ----------
    r : int
        The radius of the desired disk
    smooth : boolean
        If ``True`` (default) then the disk will not have the litte
        nibs on the surfaces.

    Returns
    -------
    disk : ndarray
        A numpy array of 1 and 0 suitable for use as a structuring element
    """
    s = np.zeros((2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            if ((i - r)**2 + (j - r)**2)**0.5 <= thresh:
                s[i, j] = 1
    return s


@njit(parallel=False)
def _make_ball(r, smooth=True):  # pragma: no cover
    r"""
    Generate a spherical structuring element of the given radius

    Parameters
    ----------
    r : int
        The radius of the desired ball
    smooth : boolean
        If ``True`` (default) then the ball will not have the litte
        nibs on the surfaces.

    Returns
    -------
    ball : ndarray
        A numpy array of 1 and 0 suitable for use as a structuring element
    """
    s = np.zeros((2*r+1, 2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            for k in range(2*r+1):
                if ((i - r)**2 + (j - r)**2 + (k - r)**2)**0.5 <= thresh:
                    s[i, j, k] = 1
    return s



r"""

Collection of functions for altering images based on structural properties
##########################################################################

This module contains a variety of functions for altering images based on
the structural characteristics, such as pore sizes.  A definition of a
*filter* is a function that returns an image the shape as the original
image, but with altered values.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    filters.apply_chords
    filters.apply_chords_3D
    filters.apply_padded
    filters.chunked_func
    filters.distance_transform_lin
    filters.fftmorphology
    filters.fill_blind_pores
    filters.find_disconnected_voxels
    filters.find_dt_artifacts
    filters.find_peaks
    filters.find_trapped_regions
    filters.flood
    filters.flood_func
    filters.hold_peaks
    filters.ibip
    filters.ibip_gpu
    filters.imagej
    filters.local_thickness
    filters.nl_means_layered
    filters.nphase_border
    filters.pc_to_satn
    filters.porosimetry
    filters.prune_branches
    filters.reduce_peaks
    filters.region_size
    filters.satn_to_seq
    filters.seq_to_satn
    filters.size_to_satn
    filters.size_to_seq
    filters.snow_partitioning
    filters.snow_partitioning_n
    filters.snow_partitioning_parallel
    filters.trim_disconnected_blobs
    filters.trim_extrema
    filters.trim_floating_solid
    filters.trim_nearby_peaks
    filters.trim_nonpercolating_paths
    filters.trim_saddle_points
    filters.trim_small_clusters

"""

from ._funcs import *
from ._snows import *
from ._size_seq_satn import *
from ._nlmeans import *
from ._fftmorphology import *
from . import imagej



r"""

Collection of functions for generating synthetic images
#######################################################

This module contains a variety of functions for generating artificial
images of porous materials, generally for testing, validation,
debugging, and illustration purposes.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    generators.rsa
    generators.random_spheres
    generators.blobs
    generators.borders
    generators.bundle_of_tubes
    generators.cylinders
    generators.cylindrical_plug
    generators.faces
    generators.fractal_noise
    generators.insert_shape
    generators.lattice_spheres
    generators.line_segment
    generators.overlapping_spheres
    generators.polydisperse_spheres
    generators.pseudo_electrostatic_packing
    generators.pseudo_gravity_packing
    generators.random_cantor_dust
    generators.sierpinski_foam
    generators.voronoi_edges

"""

from ._imgen import *
from ._pseudo_packings import *
from ._noise import *
from ._borders import *
from ._fractals import *
from ._spheres_from_coords import *



import numpy as np
from edt import edt
from porespy.tools import get_tqdm
from porespy.tools import get_border, make_contiguous
from porespy.tools import _insert_disk_at_points
from porespy.tools import Results
from numba import njit
from porespy import settings
tqdm = get_tqdm()


__all__ = [
    "ibip",
]


def ibip(im, inlets=None, dt=None, maxiter=10000):
    r"""
    Performs invasion percolation on given image using iterative image dilation

    Parameters
    ----------
    im : ND-array
        Boolean array with ``True`` values indicating void voxels
    inlets : ND-array
        Boolean array with ``True`` values indicating where the invading fluid
        is injected from.  If ``None``, all faces will be used.
    dt : ND-array (optional)
        The distance transform of ``im``.  If not provided it will be
        calculated, so supplying it saves time.
    maxiter : scalar
        The number of steps to apply before stopping.  The default is to run
        for 10,000 steps which is almost certain to reach completion if the
        image is smaller than about 250-cubed.

    Returns
    -------
    results : Results object
        A custom object with the following two arrays as attributes:

        'inv_sequence'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the sequence at which it was invaded.

        'inv_size'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the ``inv_size`` at which was filled.

    See Also
    --------
    porosimetry

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/reference/ibip.html>`_
    to view online example.

    """
    # Process the boundary image
    if inlets is None:
        inlets = get_border(shape=im.shape, mode='faces')
    bd = np.copy(inlets > 0)
    if dt is None:  # Find dt if not given
        dt = edt(im)
    dt = dt.astype(int)  # Conert the dt to nearest integer
    # Initialize inv image with -1 in the solid, and 0's in the void
    inv = -1*(~im)
    sizes = -1*(~im)
    scratch = np.copy(bd)
    for step in tqdm(range(1, maxiter), **settings.tqdm):
        pt = _where(bd)
        scratch = np.copy(bd)
        temp = _insert_disk_at_points(im=scratch, coords=pt,
                                       r=1, v=1, smooth=False)
        # Reduce to only the 'new' boundary
        edge = temp*(dt > 0)
        if ~edge.any():
            break
        # Find the maximum value of the dt underlaying the new edge
        r_max = (dt*edge).max()
        # Find all values of the dt with that size
        dt_thresh = dt >= r_max
        # Extract the actual coordinates of the insertion sites
        pt = _where(edge*dt_thresh)
        inv = _insert_disk_at_points(im=inv, coords=pt,
                                      r=r_max, v=step, smooth=True)
        sizes = _insert_disk_at_points(im=sizes, coords=pt,
                                        r=r_max, v=r_max, smooth=True)
        dt, bd = _update_dt_and_bd(dt, bd, pt)
    # Convert inv image so that uninvaded voxels are set to -1 and solid to 0
    temp = inv == 0  # Uninvaded voxels are set to -1 after _ibip
    inv[~im] = 0
    inv[temp] = -1
    inv = make_contiguous(im=inv, mode='symmetric')
    # Deal with invasion sizes similarly
    temp = sizes == 0
    sizes[~im] = 0
    sizes[temp] = -1
    results = Results()
    results.inv_sequence = inv
    results.inv_sizes = sizes
    return results


@njit(parallel=False)
def _where(arr):
    inds = np.where(arr)
    result = np.vstack(inds)
    return result


@njit()
def _update_dt_and_bd(dt, bd, pt):
    if dt.ndim == 2:
        for i in range(pt.shape[1]):
            bd[pt[0, i], pt[1, i]] = True
            dt[pt[0, i], pt[1, i]] = 0
    else:
        for i in range(pt.shape[1]):
            bd[pt[0, i], pt[1, i], pt[2, i]] = True
            dt[pt[0, i], pt[1, i], pt[2, i]] = 0
    return dt, bd


if __name__ == "__main__":
    import numpy as np
    import porespy as ps
    import matplotlib.pyplot as plt
    from copy import copy

    # %% Run this cell to regenerate the variables in drainage
    np.random.seed(6)
    bg = 'white'
    plots = True
    im = ps.generators.blobs(shape=[300, 300], porosity=0.7, blobiness=2)
    inlets = np.zeros_like(im)
    inlets[0, :] = True
    ip = ps.simulations.ibip(im=im, inlets=inlets)

    # %% Generate some plots
    if plots:
        cmap = copy(plt.cm.plasma)
        cmap.set_under(color='black')
        cmap.set_over(color='grey')
        fig, ax = plt.subplots(1, 1)
        kw = ps.visualization.prep_for_imshow(ip.inv_sequence, im)
        kw['vmin'] = 0
        ax.imshow(**kw, cmap=cmap)
